############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
from datetime import datetime as dt
import gc
# external packages
import PyQt5
import numpy as np
from matplotlib import ticker
import matplotlib.pyplot as plt
# local import
from mw4.base import tpool
from mw4.gui import widget
from mw4.gui.widgets import measure_ui


class MeasureWindow(widget.MWidget):
    """
    the measure window class handles

    """

    __all__ = ['MeasureWindow',
               ]
    version = '0.5'
    logger = logging.getLogger(__name__)

    NUMBER_POINTS = 250
    NUMBER_XTICKS = 5

    def __init__(self, app):
        super().__init__()
        self.app = app

        self.ui = measure_ui.Ui_MeasureDialog()
        self.ui.setupUi(self)
        self.initUI()

        self.refreshCounter = 1
        self.measureIndex = 0
        self.timeIndex = 0

        self.mSetUI = [self.ui.measureSet1,
                       self.ui.measureSet2,
                       self.ui.measureSet3,
                       ]

        self.plotFunc = {'None': None,
                         'RA Stability': self.plotRa,
                         'DEC Stability': self.plotDec,
                         'Temperature': self.plotTemperature,
                         'Pressure': self.plotPressure,
                         'Humidity': self.plotHumidity,
                         'Sky Quality': self.plotSQR,
                         'Voltage': self.plotVoltage,
                         'Current': self.plotCurrent,
                         }

        self.timeScale = {'  1s Tick   -  4 min': 1,
                          '  2s Ticks  -  8 min': 2,
                          '  4s Ticks  - 16 min': 4,
                          '  8s Ticks  - 32 min': 8,
                          ' 16s Ticks  -  1 hour': 16,
                          ' 32s Ticks  -  2 hours': 32,
                          ' 64s Ticks  -  4 hours': 64,
                          '128s Ticks  -  9 hours': 128,
                          }

        # doing the matplotlib embedding
        self.measureMat = self.embedMatplot(self.ui.measure)
        self.measureMat.parentWidget().setStyleSheet(self.BACK_BG)

        self.initConfig()
        self.showWindow()

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        if 'measureW' not in self.app.config:
            self.app.config['measureW'] = {}
        config = self.app.config['measureW']
        x = config.get('winPosX', 100)
        y = config.get('winPosY', 100)
        if x > self.screenSizeX:
            x = 0
        if y > self.screenSizeY:
            y = 0
        self.move(x, y)
        height = config.get('height', 600)
        width = config.get('width', 800)
        self.resize(width, height)
        self.setupButtons()
        self.ui.measureSet1.setCurrentIndex(config.get('measureSet1', 0))
        self.ui.measureSet2.setCurrentIndex(config.get('measureSet2', 0))
        self.ui.measureSet3.setCurrentIndex(config.get('measureSet3', 0))
        self.ui.timeSet.setCurrentIndex(config.get('timeSet', 0))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['measureW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()
        config['width'] = self.width()
        config['measureSet1'] = self.ui.measureSet1.currentIndex()
        config['measureSet2'] = self.ui.measureSet2.currentIndex()
        config['measureSet3'] = self.ui.measureSet3.currentIndex()
        config['timeSet'] = self.ui.timeSet.currentIndex()

        return True

    def showWindow(self):
        """

        :return:
        """
        self.drawMeasure()
        self.show()

        # signals for gui
        self.ui.timeSet.currentIndexChanged.connect(self.setCycleRefresh)
        self.ui.measureSet1.currentIndexChanged.connect(self.cycleRefresh)
        self.ui.measureSet2.currentIndexChanged.connect(self.cycleRefresh)
        self.ui.measureSet3.currentIndexChanged.connect(self.cycleRefresh)
        self.app.update1s.connect(self.cycleRefresh)
        return True

    def closeEvent(self, closeEvent):
        """

        :param closeEvent:
        :return:
        """

        self.storeConfig()
        # signals for gui
        self.ui.timeSet.currentIndexChanged.disconnect(self.setCycleRefresh)
        self.ui.measureSet1.currentIndexChanged.disconnect(self.cycleRefresh)
        self.ui.measureSet2.currentIndexChanged.disconnect(self.cycleRefresh)
        self.ui.measureSet3.currentIndexChanged.disconnect(self.cycleRefresh)
        self.app.update1s.disconnect(self.cycleRefresh)

        plt.close(self.measureMat.figure)

        super().closeEvent(closeEvent)

    def setupButtons(self):
        """
        setupButtons prepares the dynamic content od the buttons in measurement window. it
        write the bottom texts and number as well as the coloring for the actual setting

        :return: success for test purpose
        """

        for mSet in self.mSetUI:
            mSet.clear()
            mSet.setView(PyQt5.QtWidgets.QListView())
            for text in self.plotFunc.keys():
                mSet.addItem(text)

        tSet = self.ui.timeSet
        tSet.clear()
        tSet.setView(PyQt5.QtWidgets.QListView())
        for text in self.timeScale.keys():
            tSet.addItem(text)

        return True

    def setCycleRefresh(self):
        """

        :return: True for test purpose
        """

        self.refreshCounter = 0
        self.cycleRefresh()

        return True

    def cycleRefresh(self):
        """

        :return: True for test purpose
        """

        cycle = self.timeScale[self.ui.timeSet.currentText()]
        if not self.refreshCounter % cycle:
            self.drawMeasure(cycle=cycle)
        self.refreshCounter += 1

        return True

    def setupAxes(self, figure=None, numberPlots=3):
        """
        setupAxes cleans up the axes object in figure an setup a new plotting. it draws
        grid, ticks etc.

        :param numberPlots:
        :param figure: axes object of figure
        :return:
        """

        if figure is None:
            return None
        if numberPlots > 3:
            return None
        if numberPlots < 0:
            return None

        for axe in figure.axes:
            axe.cla()
            del axe
            gc.collect()

        figure.clf()
        figure.subplots_adjust(left=0.1, right=0.95, bottom=0.05, top=0.95)
        for i in range(numberPlots):
            self.measureMat.figure.add_subplot(numberPlots, 1, i + 1, facecolor=None)

        for axe in figure.axes:
            axe.set_facecolor((0, 0, 0, 0))
            axe.tick_params(colors=self.M_BLUE,
                            labelsize=12)
            axe.spines['bottom'].set_color(self.M_BLUE)
            axe.spines['top'].set_color(self.M_BLUE)
            axe.spines['left'].set_color(self.M_BLUE)
            axe.spines['right'].set_color(self.M_BLUE)

        return figure.axes

    def plotRa(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'delta RA [arcsec]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        axe.plot(data['time'][start:-1:cycle],
                 data['raJNow'][start:-1:cycle],
                 marker='o',
                 markersize=1,
                 color=self.M_WHITE,
                 )
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.set_ylim(-0.4, 0.4)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def plotDec(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'delta DEC [arcsec]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        axe.plot(data['time'][start:-1:cycle],
                 data['decJNow'][start:-1:cycle],
                 marker='o',
                 markersize=1,
                 color=self.M_WHITE,
                 )
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.set_ylim(-4, 4)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def plotTemperature(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Temperature [°C]'
        start = -self.NUMBER_POINTS * cycle

        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        r1, = axe.plot(data['time'][start:-1:cycle],
                       data['envTemp'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_WHITE,
                       )
        r2, = axe.plot(data['time'][start:-1:cycle],
                       data['powTemp'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_PINK,
                       )
        r3, = axe.plot(data['time'][start:-1:cycle],
                       data['skyTemp'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_YELLOW,
                       )
        r4, = axe.plot(data['time'][start:-1:cycle],
                       data['envDew'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_GREEN,
                       )
        r5, = axe.plot(data['time'][start:-1:cycle],
                       data['powDew'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_RED,
                       )
        legend = axe.legend([r1, r2, r3, r4, r5],
                            ['Env Temp', 'Pow Temp', 'Sky Temp', 'Env Dew', 'Pow Dew'],
                            facecolor=self.M_BLACK,
                            edgecolor=self.M_BLUE,
                            )
        for text in legend.get_texts():
            text.set_color(self.M_BLUE)
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.margins(y=0.2)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def plotPressure(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Pressure [hPas]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        axe.plot(data['time'][start:-1:cycle],
                 data['envPress'][start:-1:cycle],
                 marker='o',
                 markersize=1,
                 color=self.M_WHITE,
                 )
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.margins(y=0.2)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.0f',
                                                                      ))
        return True

    def plotHumidity(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Humidity [%]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        r1, = axe.plot(data['time'][start:-1:cycle],
                       data['envHum'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_WHITE,
                       )
        r2, = axe.plot(data['time'][start:-1:cycle],
                       data['powHum'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_PINK,
                       )
        legend = axe.legend([r1, r2],
                            ['Env Hum', 'Pow Hum'],
                            facecolor=self.M_BLACK,
                            edgecolor=self.M_BLUE,
                            )
        for text in legend.get_texts():
            text.set_color(self.M_BLUE)
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.set_ylim(-0, 100)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.0f',
                                                                      ))
        return True

    def plotSQR(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Sky Quality [mpas]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        axe.plot(data['time'][start:-1:cycle],
                 data['skySQR'][start:-1:cycle],
                 marker='o',
                 markersize=1,
                 color=self.M_WHITE,
                 )
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.margins(y=0.2)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def plotVoltage(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Power Voltage on Rack [V]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        axe.plot(data['time'][start:-1:cycle],
                 data['powVolt'][start:-1:cycle],
                 marker='o',
                 markersize=1,
                 color=self.M_WHITE,
                 )
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.margins(y=0.2)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def plotCurrent(self, axe=None, title='', data=None, cycle=None):
        """
        drawRaDec show the specific graph for plotting the ra dec deviations. this
        is done with two color and axes to distinguish the values for ra and dec.
        ideally the values are around zero, but the scales of ra and dec axis have to be
        different.

        :param axe: axe for plotting
        :param title: title text
        :param data: data location
        :param cycle: cycle time for measurement
        :return: success
        """
        ylabel = 'Power Current [A]'
        start = -self.NUMBER_POINTS * cycle
        axe.set_title(title,
                      color=self.M_BLUE,
                      fontweight='bold',
                      fontsize=16)
        axe.set_ylabel(ylabel,
                       color=self.M_BLUE,
                       fontweight='bold',
                       fontsize=12)
        r1, = axe.plot(data['time'][start:-1:cycle],
                       data['powCurr'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_WHITE,
                       )
        r2, = axe.plot(data['time'][start:-1:cycle],
                       data['powCurr1'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_PINK,
                       )
        r3, = axe.plot(data['time'][start:-1:cycle],
                       data['powCurr2'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_YELLOW,
                       )
        r4, = axe.plot(data['time'][start:-1:cycle],
                       data['powCurr3'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_GREEN,
                       )
        r5, = axe.plot(data['time'][start:-1:cycle],
                       data['powCurr4'][start:-1:cycle],
                       marker='o',
                       markersize=1,
                       color=self.M_RED,
                       )
        legend = axe.legend([r1, r2, r3, r4, r5],
                            ['Curr Sum', 'Curr P1', 'Curr P2', 'Curr P3', 'Curr P4'],
                            facecolor=self.M_BLACK,
                            edgecolor=self.M_BLUE,
                            )
        for text in legend.get_texts():
            text.set_color(self.M_BLUE)
        axe.grid(True, color=self.M_GREY, alpha=1)
        axe.margins(y=0.2)
        axe.get_yaxis().set_major_locator(ticker.MaxNLocator(nbins=8,
                                                             integer=True,
                                                             min_n_ticks=4,
                                                             prune='both',
                                                             ))
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.1f',
                                                                      ))
        return True

    def drawMeasure(self, cycle=1):
        """
        drawMeasure does the basic preparation for making the plot. it checks for borders
        and does finally the content dispatcher. currently there is no chance to implement
        a basic pattern as the graphs differ heavily.

        :param cycle:
        :return: success
        """

        data = self.app.measure.data

        if len(data['time']) < 4:
            return False

        numberPlots = 0
        for mSet in self.mSetUI:
            if mSet.currentText() == 'None':
                continue
            numberPlots += 1

        axes = self.setupAxes(figure=self.measureMat.figure, numberPlots=numberPlots)

        if axes is None:
            return False
        if not numberPlots:
            self.measureMat.figure.canvas.draw()
            return False

        grid = int(self.NUMBER_POINTS / self.NUMBER_XTICKS)
        ratio = cycle * grid
        time_end = data['time'][-1]

        time_ticks = np.arange(-self.NUMBER_XTICKS, 1, 1)
        time_ticks = time_ticks * ratio * 1000000
        time_ticks = time_ticks + time_end
        time_labels = [x.astype(dt).strftime('%H:%M:%S') for x in time_ticks]

        for i, axe in enumerate(axes):
            axe.set_xticks(time_ticks)
            axe.set_xlim(time_ticks[0], time_ticks[-1])
            if i == len(axes) - 1:
                axe.set_xticklabels(time_labels)
            else:
                axe.set_xticklabels([])

        for axe, mSet in zip(axes, self.mSetUI):
            key = mSet.currentText()
            if self.plotFunc[key] is None:
                continue
            self.plotFunc[key](axe=axe,
                               title=mSet.currentText(),
                               data=data,
                               cycle=cycle)
            axe.figure.canvas.draw()

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
# external packages
# local import


class SettHorizon(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):

        self.ui.saveHorizonMask.clicked.connect(self.saveHorizonMask)
        self.ui.saveHorizonMaskAs.clicked.connect(self.saveHorizonMaskAs)
        self.ui.loadHorizonMask.clicked.connect(self.loadHorizonMask)
        self.ui.checkAutoDeletePoints.clicked.connect(self.autoDeletePoints)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        self.ui.horizonFileName.setText(config.get('horizonFileName', ''))
        self.ui.checkAutoDeletePoints.setChecked(config.get('checkAutoDeletePoints', False))
        self.ui.checkSortNothing.setChecked(config.get('checkSortNothing', True))
        self.ui.checkSortEW.setChecked(config.get('checkSortEW', False))
        self.ui.checkSortHL.setChecked(config.get('checkSortHL', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['horizonFileName'] = self.ui.horizonFileName.text()
        config['checkAutoDeletePoints'] = self.ui.checkAutoDeletePoints.isChecked()
        config['checkSortNothing'] = self.ui.checkSortNothing.isChecked()
        config['checkSortEW'] = self.ui.checkSortEW.isChecked()
        config['checkSortHL'] = self.ui.checkSortHL.isChecked()

        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """
        return True

    def loadHorizonMask(self):
        """
        loadHorizonMask calls a file selector box and selects the filename to be loaded

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        loadFilePath, fileName, ext = self.openFile(self,
                                                    'Open horizon mask file',
                                                    folder,
                                                    'Horizon mask files (*.hpts)',
                                                    )
        if not loadFilePath:
            return False
        suc = self.app.data.loadHorizonP(fileName=fileName)
        if suc:
            self.ui.horizonFileName.setText(fileName)
            self.app.message.emit('Horizon mask [{0}] loaded'.format(fileName), 0)
            self.app.hemisphereW.drawHemisphere()
        else:
            self.app.message.emit('Horizon mask [{0}] cannot no be loaded'
                                  .format(fileName), 2)
        return True

    def saveHorizonMask(self):
        """
        saveHorizonMask calls saving the build file

        :return: success
        """

        fileName = self.ui.horizonFileName.text()
        if not fileName:
            self.app.message.emit('Horizon mask file name not given', 2)
            return False
        suc = self.app.data.saveHorizonP(fileName=fileName)
        if suc:
            self.app.message.emit('Horizon mask [{0}] saved'.format(fileName), 0)
        else:
            self.app.message.emit('Horizon mask [{0}] cannot no be saved'
                                  .format(fileName), 2)
        return True

    def saveHorizonMaskAs(self):
        """
        saveHorizonMaskAs calls a file selector box and selects the filename to be save

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        saveFilePath, fileName, ext = self.saveFile(self,
                                                    'Save horizon mask file',
                                                    folder,
                                                    'Horizon mask files (*.hpts)',
                                                    )
        if not saveFilePath:
            return False
        suc = self.app.data.saveHorizonP(fileName=fileName)
        if suc:
            self.ui.horizonFileName.setText(fileName)
            self.app.message.emit('Horizon mask [{0}] saved'.format(fileName), 0)
        else:
            self.app.message.emit('Horizon mask [{0}] cannot no be saved'
                                  .format(fileName), 2)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
# external packages
import PyQt5.QtCore
import PyQt5.QtWidgets
import PyQt5.uic
import numpy as np
import matplotlib.pyplot
from mountcontrol import convert
# local import


class ManageModel(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self.runningTargetRMS = False
        ms = self.app.mount.signals
        ms.alignDone.connect(self.showModelPolar)
        ms.namesDone.connect(self.setNameList)

        self.ui.checkShowErrorValues.stateChanged.connect(self.showModelPolar)
        self.ui.refreshName.clicked.connect(self.refreshName)
        self.ui.refreshModel.clicked.connect(self.refreshModel)
        self.ui.clearModel.clicked.connect(self.clearModel)
        self.ui.loadName.clicked.connect(self.loadName)
        self.ui.saveName.clicked.connect(self.saveName)
        self.ui.deleteName.clicked.connect(self.deleteName)
        self.ui.runTargetRMS.clicked.connect(self.runTargetRMS)
        self.ui.cancelTargetRMS.clicked.connect(self.cancelTargetRMS)
        self.ui.deleteWorstPoint.clicked.connect(self.deleteWorstPoint)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        self.ui.checkShowErrorValues.setChecked(config.get('checkShowErrorValues', False))
        self.ui.targetRMS.setValue(config.get('targetRMS', 99))
        self.showModelPolar(None)
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['checkShowErrorValues'] = self.ui.checkShowErrorValues.isChecked()
        config['targetRMS'] = self.ui.targetRMS.value()
        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """
        self.wIcon(self.ui.runTargetRMS, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)
        self.wIcon(self.ui.cancelTargetRMS, PyQt5.QtWidgets.QStyle.SP_DialogCancelButton)
        self.wIcon(self.ui.loadName, PyQt5.QtWidgets.QStyle.SP_DirOpenIcon)
        self.wIcon(self.ui.saveName, PyQt5.QtWidgets.QStyle.SP_DialogSaveButton)
        self.wIcon(self.ui.deleteName, PyQt5.QtWidgets.QStyle.SP_TrashIcon)
        self.wIcon(self.ui.refreshName, PyQt5.QtWidgets.QStyle.SP_BrowserReload)
        self.wIcon(self.ui.refreshModel, PyQt5.QtWidgets.QStyle.SP_BrowserReload)
        return True

    def setNameList(self, model):
        """
        setNameList populates the list of model names in the main window. before adding the
        data, the existent list will be deleted.

        :return:    True if ok for testing
        """

        self.ui.nameList.clear()
        for name in model.nameList:
            self.ui.nameList.addItem(name)
        self.ui.nameList.sortItems()
        self.ui.nameList.update()
        return True

    def showModelPolar(self, model):
        """
        showModelPolar draws a polar plot of the align model stars and their errors in
        color. the basic setup of the plot is taking place in the central widget class.
        which is instantiated from there. important: the coordinate in model is in HA and
        DEC  and not in RA and DEC. using skyfield is a little bit misleading, because you
        address the hour angle as .ra.hours

        :return:    True if ok for testing
        """

        # check entry conditions for displaying a polar plot
        if model is None:
            hasNoStars = True
        elif not hasattr(model, 'starList'):
            hasNoStars = True
        else:
            hasNoStars = model.starList is None or not model.starList

        if hasNoStars:
            # clear the plot and return
            fig, axes = self.clearPolar(self.polarPlot)
            # used constrained_layout = True instead
            # fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.85)
            axes.figure.canvas.draw()
            return False

        # start with plotting
        fig, axes = self.clearPolar(self.polarPlot)
        lat = self.app.config.get('topoLat', 51.47)

        altitude = []
        azimuth = []
        error = []
        for star in model.starList:
            alt, az = convert.topoToAltAz(star.coord.ra.hours,
                                          star.coord.dec.degrees,
                                          lat)
            altitude.append(alt)
            azimuth.append(az)
            error.append(star.errorRMS)
        altitude = np.asarray(altitude)
        azimuth = np.asarray(azimuth)
        error = np.asarray(error)

        # and plot it
        cm = matplotlib.pyplot.cm.get_cmap('RdYlGn_r')
        colors = np.asarray(error)
        scaleErrorMax = max(colors)
        scaleErrorMin = min(colors)
        area = [200 if x >= max(colors) else 60 for x in error]
        theta = azimuth / 180.0 * np.pi
        r = 90 - altitude
        scatter = axes.scatter(theta,
                               r,
                               c=colors,
                               vmin=scaleErrorMin,
                               vmax=scaleErrorMax,
                               s=area,
                               cmap=cm,
                               zorder=0,
                               )
        if self.ui.checkShowErrorValues.isChecked():
            for star in model.starList:
                text = '{0:3.1f}'.format(star.errorRMS)
                axes.annotate(text,
                              xy=(theta[star.number - 1],
                                  r[star.number - 1]),
                              color='#2090C0',
                              fontsize=9,
                              fontweight='bold',
                              zorder=1,
                              )
        formatString = matplotlib.ticker.FormatStrFormatter('%1.0f')
        colorbar = fig.colorbar(scatter,
                                pad=0.1,
                                fraction=0.12,
                                aspect=25,
                                shrink=0.9,
                                format=formatString,
                                )
        colorbar.set_label('Error [arcsec]', color='white')
        yTicks = matplotlib.pyplot.getp(colorbar.ax.axes, 'yticklabels')
        matplotlib.pyplot.setp(yTicks,
                               color='#2090C0',
                               fontweight='bold')
        axes.set_rmax(90)
        axes.set_rmin(0)
        axes.figure.canvas.draw()
        return True

    def clearRefreshName(self):
        """
        clearRefreshName is the buddy function for refreshName

        :return: True for test purpose
        """

        self.changeStyleDynamic(self.ui.refreshName, 'running', 'false')
        self.ui.deleteName.setEnabled(True)
        self.ui.saveName.setEnabled(True)
        self.ui.loadName.setEnabled(True)
        self.app.mount.signals.namesDone.disconnect(self.clearRefreshName)
        self.app.message.emit('Model names refreshed', 0)
        return True

    def refreshName(self):
        """
        refreshName disables interfering functions in gui and start reloading the
        names list for model in the mount computer. it connects a link to clearRefreshNames
        which enables the former disabled gui buttons and removes the link to the method.
        after it triggers the refresh of names, it finished, because behaviour is event
        driven

        :return: True for test purpose
        """

        self.app.mount.signals.namesDone.connect(self.clearRefreshName)
        self.ui.deleteName.setEnabled(False)
        self.ui.saveName.setEnabled(False)
        self.ui.loadName.setEnabled(False)
        self.changeStyleDynamic(self.ui.refreshName, 'running', 'true')
        self.app.mount.getNames()
        return True

    def loadName(self):
        """
        loadName take the given name and loads the stored model as the actual alignment
        model for the mount. after that it refreshes the alignment model data in
        mountwizzard

        :return: success
        """

        if self.ui.nameList.currentItem() is None:
            self.app.message.emit('No model name selected', 2)
            return False
        modelName = self.ui.nameList.currentItem().text()
        suc = self.app.mount.model.loadName(modelName)
        if not suc:
            self.app.message.emit('Model [{0}] cannot be loaded'.format(modelName), 2)
            return False
        else:
            self.app.message.emit('Model [{0}] loaded'.format(modelName), 0)
            self.refreshModel()
            return True

    def saveName(self):
        """
        saveName take the given name and saves the actual alignment model to the model
        database in the mount computer. after that it refreshes the list of the alignment
        model names in mountwizzard.

        :return: success
        """

        dlg = PyQt5.QtWidgets.QInputDialog()
        modelName, ok = dlg.getText(self,
                                    'Save model',
                                    'New model name',
                                    PyQt5.QtWidgets.QLineEdit.Normal,
                                    '',
                                    )
        if modelName is None or not modelName:
            self.app.message.emit('No model name given', 2)
            return False
        if not ok:
            return False
        suc = self.app.mount.model.storeName(modelName)
        if not suc:
            self.app.message.emit('Model [{0}] cannot be saved'.format(modelName), 2)
            return False
        else:
            self.app.message.emit('Model [{0}] saved'.format(modelName), 0)
            self.refreshName()
            return True

    def deleteName(self):
        """
        deleteName take the given name and deletes it from the model database in the
        mount computer. after that it refreshes the list of the alignment model names in
        mountwizzard.

        :return: success
        """

        if self.ui.nameList.currentItem() is None:
            self.app.message.emit('No model name selected', 2)
            return False
        modelName = self.ui.nameList.currentItem().text()
        msg = PyQt5.QtWidgets.QMessageBox
        reply = msg.question(self,
                             'Delete model',
                             'Delete model from database?',
                             msg.Yes | msg.No,
                             msg.No,
                             )
        if reply != msg.Yes:
            return False
        suc = self.app.mount.model.deleteName(modelName)
        if not suc:
            self.app.message.emit('Model [{0}] cannot be deleted'.format(modelName), 2)
            return False
        else:
            self.app.message.emit('Model [{0}] deleted'.format(modelName), 0)
            self.refreshName()
            return True

    def clearRefreshModel(self):
        """
        clearRefreshModel is the buddy function for refreshModel

        :return: True for test purpose
        """

        self.changeStyleDynamic(self.ui.refreshModel, 'running', 'false')
        self.ui.deleteWorstPoint.setEnabled(True)
        self.ui.runTargetRMS.setEnabled(True)
        self.ui.cancelTargetRMS.setEnabled(True)
        self.ui.clearModel.setEnabled(True)
        self.app.mount.signals.alignDone.disconnect(self.clearRefreshModel)
        self.app.message.emit('Align model data refreshed', 0)
        return True

    def refreshModel(self):
        """
        refreshModel disables interfering functions in gui and start reloading the
        alignment model from the mount computer. it connects a link to clearRefreshModel
        which enables the former disabled gui buttons and removes the link to the method.
        after it triggers the refresh of names, it finished, because behaviour is event
        driven

        :return: True for test purpose
        """

        self.changeStyleDynamic(self.ui.refreshModel, 'running', 'true')
        self.app.mount.signals.alignDone.connect(self.clearRefreshModel)
        self.ui.deleteWorstPoint.setEnabled(False)
        self.ui.runTargetRMS.setEnabled(False)
        self.ui.cancelTargetRMS.setEnabled(False)
        self.ui.clearModel.setEnabled(False)
        self.app.mount.getAlign()

        return True

    def clearRunTargetRMS(self):
        """
        clearRunTargetRMS is the buddy function for runTargetRMS

        :return: True for test purpose
        """

        mount = self.app.mount
        if mount.model.errorRMS < self.ui.targetRMS.value():
            self.runningTargetRMS = False

        if self.runningTargetRMS:
            wIndex = mount.model.starList.index(max(mount.model.starList))
            wStar = mount.model.starList[wIndex]
            suc = mount.model.deletePoint(wStar.number)
            if not suc:
                self.runningTargetRMS = False
                self.app.message.emit(f'Star [{wStar.number}] cannot be deleted', 2)
            else:
                text = f'Star [{wStar.number:02d}]: RMS of [{wStar.errorRMS:04.1f}] deleted'
                self.app.message.emit(text, 0)
            mount.getAlign()
        else:
            self.changeStyleDynamic(self.ui.runTargetRMS, 'running', 'false')
            self.ui.deleteWorstPoint.setEnabled(True)
            self.ui.clearModel.setEnabled(True)
            self.ui.refreshModel.setEnabled(True)
            mount.signals.alignDone.disconnect(self.clearRunTargetRMS)
            self.changeStyleDynamic(self.ui.cancelTargetRMS, 'cancel', 'false')
            self.app.message.emit('Optimizing done', 0)
        return True

    def runTargetRMS(self):
        """
        refreshModel disables interfering functions in gui and start reloading the
        alignment model from the mount computer. it connects a link to clearRefreshModel
        which enables the former disabled gui buttons and removes the link to the method.
        after it triggers the refresh of names, it finished, because behaviour is event
        driven

        :return: True for test purpose
        """

        self.app.message.emit('Start optimizing model', 0)
        self.runningTargetRMS = True
        self.app.mount.signals.alignDone.connect(self.clearRunTargetRMS)
        self.ui.deleteWorstPoint.setEnabled(False)
        self.ui.clearModel.setEnabled(False)
        self.ui.refreshModel.setEnabled(False)
        self.changeStyleDynamic(self.ui.runTargetRMS, 'running', 'true')
        self.clearRunTargetRMS()

        return True

    def cancelTargetRMS(self):
        """
        cancelTargetRMS just resets the running Flag for RMS target optimization -> enables
        the stop of optimization.

        :return:
        """

        self.runningTargetRMS = False
        self.changeStyleDynamic(self.ui.cancelTargetRMS, 'cancel', 'true')
        return True

    def clearModel(self):
        """
        clearModel removes the actual alignment model.

        :return:
        """

        msg = PyQt5.QtWidgets.QMessageBox
        reply = msg.question(self,
                             'Clear model',
                             'Clear actual alignment model?',
                             msg.Yes | msg.No,
                             msg.No,
                             )
        if reply == msg.No:
            return False
        suc = self.app.mount.model.clearAlign()
        if not suc:
            self.app.message.emit('Actual model cannot be cleared', 2)
            return False
        else:
            self.app.message.emit('Actual model cleared', 0)
            self.refreshModel()
            return True

    def deleteWorstPoint(self):
        """
        deleteWorstPoint selects from the actual model stored in the mount the highest
        value of error, get it's index and deletes the point. afterward it refreshes the
        gui

        :return:
        """

        model = self.app.mount.model
        wIndex = model.starList.index(max(model.starList))
        wStar = model.starList[wIndex]

        suc = model.deletePoint(wStar.number)

        if not suc:
            self.app.message.emit('Worst point cannot be deleted', 2)
            return False
        else:
            self.app.message.emit('Worst point deleted', 0)
            self.refreshModel()
            return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
# external packages
import PyQt5
# local import


class SettIndi(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self.deviceNameDropDowns = [self.ui.imagingDeviceName,
                                    self.ui.domeDeviceName,
                                    self.ui.environDeviceName,
                                    self.ui.skymeterDeviceName,
                                    self.ui.powerDeviceName,
                                    ]
        self.deviceNameDropDownKeys = ['imagingDeviceName',
                                       'domeDeviceName',
                                       'environmentDeviceName',
                                       'skymeterDeviceName',
                                       'powerDeviceName',
                                       ]

        self.app.environ.client.signals.newMessage.connect(self.indiMessage)
        self.app.skymeter.client.signals.newMessage.connect(self.indiMessage)
        self.app.power.client.signals.newMessage.connect(self.indiMessage)

        sig = self.app.dome.client.signals
        sig.serverConnected.connect(self.showIndiDomeConnected)
        sig.serverDisconnected.connect(self.showIndiDomeDisconnected)
        sig.deviceConnected.connect(self.showDomeDeviceConnected)
        sig.deviceDisconnected.connect(self.showDomeDeviceDisconnected)
        sig.newDevice.connect(self.showIndiNewDomeDevice)
        sig.removeDevice.connect(self.showIndiRemoveDomeDevice)

        sig = self.app.imaging.client.signals
        sig.serverConnected.connect(self.showIndiImagingConnected)
        sig.serverDisconnected.connect(self.showIndiImagingDisconnected)
        sig.deviceConnected.connect(self.showImagingDeviceConnected)
        sig.deviceDisconnected.connect(self.showImagingDeviceDisconnected)
        sig.newDevice.connect(self.showIndiNewImagingDevice)
        sig.removeDevice.connect(self.showIndiRemoveImagingDevice)

        sig = self.app.environ.client.signals
        sig.serverConnected.connect(self.showIndiEnvironConnected)
        sig.serverDisconnected.connect(self.showIndiEnvironDisconnected)
        sig.deviceConnected.connect(self.showEnvironDeviceConnected)
        sig.deviceDisconnected.connect(self.showEnvironDeviceDisconnected)
        sig.newDevice.connect(self.showIndiNewEnvironDevice)
        sig.removeDevice.connect(self.showIndiRemoveEnvironDevice)

        sig = self.app.skymeter.client.signals
        sig.serverConnected.connect(self.showIndiSkymeterConnected)
        sig.serverDisconnected.connect(self.showIndiSkymeterDisconnected)
        sig.deviceConnected.connect(self.showSkymeterDeviceConnected)
        sig.deviceDisconnected.connect(self.showSkymeterDeviceDisconnected)
        sig.newDevice.connect(self.showIndiNewSkymeterDevice)
        sig.removeDevice.connect(self.showIndiRemoveSkymeterDevice)

        sig = self.app.power.client.signals
        sig.serverConnected.connect(self.showIndiPowerConnected)
        sig.serverDisconnected.connect(self.showIndiPowerDisconnected)
        sig.deviceConnected.connect(self.showPowerDeviceConnected)
        sig.deviceDisconnected.connect(self.showPowerDeviceDisconnected)
        sig.newDevice.connect(self.showIndiNewPowerDevice)
        sig.removeDevice.connect(self.showIndiRemovePowerDevice)

        self.setupDeviceNameGui()
        self.ui.domeDeviceName.currentIndexChanged.connect(self.domeDispatch)
        self.ui.imagingDeviceName.currentIndexChanged.connect(self.imagingDispatch)
        self.ui.skymeterDeviceName.currentIndexChanged.connect(self.skymeterDispatch)
        self.ui.powerDeviceName.currentIndexChanged.connect(self.powerDispatch)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        for dropDown, key in zip(self.deviceNameDropDowns, self.deviceNameDropDownKeys):
            dropDown.setCurrentIndex(config.get(key, 0))

        self.ui.environHost.setText(config.get('environHost', ''))
        self.ui.environPort.setText(config.get('environPort', '7624'))
        self.ui.imagingHost.setText(config.get('imagingHost', ''))
        self.ui.imagingPort.setText(config.get('imagingPort', '7624'))
        self.ui.domeHost.setText(config.get('domeHost', ''))
        self.ui.domePort.setText(config.get('domePort', '7624'))
        self.ui.skymeterHost.setText(config.get('skymeterHost', ''))
        self.ui.skymeterPort.setText(config.get('skymeterPort', '7624'))
        self.ui.powerHost.setText(config.get('powerHost', ''))
        self.ui.powerPort.setText(config.get('powerPort', '7624'))

        self.ui.indiMessage.setChecked(config.get('indiMessage', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        for dropDown, key in zip(self.deviceNameDropDowns, self.deviceNameDropDownKeys):
            config[key] = dropDown.currentIndex()
        config['environHost'] = self.ui.environHost.text()
        config['environPort'] = self.ui.environPort.text()
        config['imagingHost'] = self.ui.imagingHost.text()
        config['imagingPort'] = self.ui.imagingPort.text()
        config['domeHost'] = self.ui.domeHost.text()
        config['domePort'] = self.ui.domePort.text()
        config['skymeterHost'] = self.ui.skymeterHost.text()
        config['skymeterPort'] = self.ui.skymeterPort.text()
        config['powerHost'] = self.ui.powerHost.text()
        config['powerPort'] = self.ui.powerPort.text()

        config['indiMessage'] = self.ui.indiMessage.isChecked()

        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """
        return True

    def setupDeviceNameGui(self):
        """
        setupRelayGui handles the dropdown lists for all devices possible in mountwizzard.
        therefore we add the necessary entries to populate the list.

        :return: success for test
        """

        for dropDown in self.deviceNameDropDowns:
            dropDown.clear()
            dropDown.setView(PyQt5.QtWidgets.QListView())
            dropDown.addItem('No device driver selected')
        # adding special items
        self.ui.imagingDeviceName.addItem('Altair')
        self.ui.imagingDeviceName.addItem('Apogee CCD')
        self.ui.imagingDeviceName.addItem('Atik CCD')
        self.ui.imagingDeviceName.addItem('CCD Simulator')
        self.ui.imagingDeviceName.addItem('Canon DSLR')
        self.ui.imagingDeviceName.addItem('DMK CCD')
        self.ui.imagingDeviceName.addItem('FLI CCD')
        self.ui.imagingDeviceName.addItem('FireFly MV')
        self.ui.imagingDeviceName.addItem('GPhoto CCD')
        self.ui.imagingDeviceName.addItem('Guide Simulator')
        self.ui.imagingDeviceName.addItem('MI CCD (ETH)')
        self.ui.imagingDeviceName.addItem('MI CCD (USB)')
        self.ui.imagingDeviceName.addItem('Meade Deep Sky Imager')
        self.ui.imagingDeviceName.addItem('Nightscape 8300 CCD')
        self.ui.imagingDeviceName.addItem('Nikon DSLR')
        self.ui.imagingDeviceName.addItem('Pentax DSLR')
        self.ui.imagingDeviceName.addItem('QHY CCD')
        self.ui.imagingDeviceName.addItem('QSI CCD')
        self.ui.imagingDeviceName.addItem('SBIG CCD')
        self.ui.imagingDeviceName.addItem('SBIG ST-I')
        self.ui.imagingDeviceName.addItem('SX CCD')
        self.ui.imagingDeviceName.addItem('Sony DSLR')
        self.ui.imagingDeviceName.addItem('Starfish CCD')
        self.ui.imagingDeviceName.addItem('ToupCam')
        self.ui.imagingDeviceName.addItem('V4L2 CCD')
        self.ui.imagingDeviceName.addItem('ZWO CCD')

        self.ui.domeDeviceName.addItem('Baader Dome')
        self.ui.domeDeviceName.addItem('Dome Scripting Gateway')
        self.ui.domeDeviceName.addItem('Dome Simulator')
        self.ui.domeDeviceName.addItem('MaxDome II')
        self.ui.domeDeviceName.addItem('NexDome')
        self.ui.domeDeviceName.addItem('RollOff Simulator')
        self.ui.domeDeviceName.addItem('ScopeDome Dome')

        self.ui.environDeviceName.addItem('AAG Cloud Watcher')
        self.ui.environDeviceName.addItem('Arduino MeteoStation')
        self.ui.environDeviceName.addItem('MBox')
        self.ui.environDeviceName.addItem('OpenWeatherMap')
        self.ui.environDeviceName.addItem('Vantage')
        self.ui.environDeviceName.addItem('Weather Meta')
        self.ui.environDeviceName.addItem('Weather Simulator')
        self.ui.environDeviceName.addItem('Weather Watcher')
        self.ui.environDeviceName.addItem('WonderGround')

        self.ui.skymeterDeviceName.addItem('SQM')

        self.ui.powerDeviceName.addItem('Pegasus UPB')

        return True

    @staticmethod
    def _removePrefix(text, prefix):
        """

        :param text:
        :param prefix:
        :return:
        """

        value = text[text.startswith(prefix) and len(prefix):]
        value = value.strip()
        return value

    def indiMessage(self, device, text):
        """
        indiMessage take a message send by indi device and puts them in the user message
        window as well.

        :param device: device name
        :param text: message received
        :return: success
        """
        if self.ui.indiMessage.isChecked():
            if text.startswith('[WARNING]'):
                text = self._removePrefix(text, '[WARNING]')
                self.app.message.emit(device + ' -> ' + text, 0)
            elif text.startswith('[ERROR]'):
                text = self._removePrefix(text, '[ERROR]')
                self.app.message.emit(device + ' -> ' + text, 2)
            else:
                self.app.message.emit(device + ' -> ' + text, 0)
            return True
        return False

    def showIndiDomeConnected(self):
        """
        showIndiDomeConnected writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit('INDI server dome connected', 0)
        return True

    def showIndiDomeDisconnected(self):
        """
        showIndiDomeDisconnected writes info to message window and recolors the status

        :return: true for test purpose
        """

        self.ui.domeDevice.setStyleSheet(self.BACK_NORM)
        self.app.message.emit('INDI server dome disconnected', 0)
        return True

    def showIndiNewDomeDevice(self, deviceName):
        """
        showIndiNewDomeDevice writes info to message window

        :return: true for test purpose
        """

        if deviceName == self.app.dome.name:
            self.app.message.emit(f'INDI dome device [{deviceName}] found', 0)
        else:
            self.app.message.emit(f'INDI dome device snoops -> [{deviceName}]', 0)

        return True

    def showIndiRemoveDomeDevice(self, deviceName):
        """
        showIndiRemoveDomeDevice writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit(f'INDI dome device [{deviceName}] removed', 0)
        return True

    def showDomeDeviceConnected(self):
        """
        showDomeDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.domeDevice.setStyleSheet(self.BACK_GREEN)
        self.deviceStat['dome'] = True
        return True

    def showDomeDeviceDisconnected(self):
        """
        showDomeDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.domeDevice.setStyleSheet(self.BACK_NORM)
        self.deviceStat['dome'] = False
        return True

    def showIndiImagingConnected(self):
        """
        showIndiImagingConnected writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit('INDI server imaging connected', 0)
        return True

    def showIndiImagingDisconnected(self):
        """
        showIndiImagingDisconnected writes info to message window and recolors the status

        :return: true for test purpose
        """

        self.ui.imagingDevice.setStyleSheet(self.BACK_NORM)
        return True

    def showIndiNewImagingDevice(self, deviceName):
        """
        showIndiNewImagingDevice writes info to message window

        :return: true for test purpose
        """

        if deviceName == self.app.imaging.name:
            self.app.message.emit(f'INDI imaging device [{deviceName}] found', 0)
        else:
            self.app.message.emit(f'INDI imaging device snoops -> [{deviceName}]', 0)

        return True

    def showIndiRemoveImagingDevice(self, deviceName):
        """
        showIndiRemoveImagingDevice writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit(f'INDI imaging device [{deviceName}] removed', 0)
        return True

    def showImagingDeviceConnected(self):
        """
        showImagingDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.imagingDevice.setStyleSheet(self.BACK_GREEN)
        self.deviceStat['imaging'] = True
        return True

    def showImagingDeviceDisconnected(self):
        """
        showImagingDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.imagingDevice.setStyleSheet(self.BACK_NORM)
        self.deviceStat['imaging'] = False
        return True

    def showIndiEnvironConnected(self):
        """
        showIndiEnvironConnected writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit('INDI server environment connected', 0)
        return True

    def showIndiEnvironDisconnected(self):
        """
        showIndiEnvironDisconnected writes info to message window and recolors the status

        :return: true for test purpose
        """

        self.ui.environDevice.setStyleSheet(self.BACK_NORM)
        self.app.message.emit('INDI server environment disconnected', 0)
        return True

    def showIndiNewEnvironDevice(self, deviceName):
        """
        showIndiNewEnvironDevice writes info to message window

        :return: true for test purpose
        """

        if deviceName == self.app.environ.name:
            self.app.message.emit(f'INDI environment device [{deviceName}] found', 0)
        else:
            self.app.message.emit(f'INDI environment device snoops -> [{deviceName}]', 0)

        return True

    def showIndiRemoveEnvironDevice(self, deviceName):
        """
        showIndiRemoveEnvironDevice writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit(f'INDI environment device [{deviceName}] removed', 0)
        return True

    def showEnvironDeviceConnected(self):
        """
        showEnvironDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.environDevice.setStyleSheet(self.BACK_GREEN)
        self.deviceStat['environment'] = True
        return True

    def showEnvironDeviceDisconnected(self):
        """
        showEnvironDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.environDevice.setStyleSheet(self.BACK_NORM)
        self.deviceStat['environment'] = False
        return True

    def showIndiSkymeterConnected(self):
        """
        showIndiSkymeterConnected writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit('INDI server skymeter connected', 0)
        return True

    def showIndiSkymeterDisconnected(self):
        """
        showIndiSkymeterDisconnected writes info to message window and recolors the status

        :return: true for test purpose
        """

        self.ui.skymeterDevice.setStyleSheet(self.BACK_NORM)
        self.app.message.emit('INDI server skymeter disconnected', 0)
        return True

    def showIndiNewSkymeterDevice(self, deviceName):
        """
        showIndiNewSkymeterDevice writes info to message window

        :return: true for test purpose
        """

        if deviceName == self.app.skymeter.name:
            self.app.message.emit(f'INDI skymeter device [{deviceName}] found', 0)
        else:
            self.app.message.emit(f'INDI skymeter device snoops -> [{deviceName}]', 0)
        return True

    def showIndiRemoveSkymeterDevice(self, deviceName):
        """
        showIndiRemoveSkymeterDevice writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit(f'INDI skymeter device [{deviceName}] removed', 0)
        return True

    def showSkymeterDeviceConnected(self):
        """
        showSkymeterDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.skymeterDevice.setStyleSheet(self.BACK_GREEN)
        self.deviceStat['skymeter'] = True
        return True

    def showSkymeterDeviceDisconnected(self):
        """
        showSkymeterDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.ui.skymeterDevice.setStyleSheet(self.BACK_NORM)
        self.deviceStat['skymeter'] = False
        return True

    def showIndiPowerConnected(self):
        """
        showIndiPowerConnected writes info to message window

        :return: true for test purpose
        """

        self.ui.powerGroup.setEnabled(True)
        self.app.message.emit('INDI server power connected', 0)
        return True

    def showIndiPowerDisconnected(self):
        """
        showIndiPowerDisconnected writes info to message window and recolors the status

        :return: true for test purpose
        """

        self.ui.powerDevice.setStyleSheet(self.BACK_NORM)
        self.app.message.emit('INDI server power disconnected', 0)
        return True

    def showIndiNewPowerDevice(self, deviceName):
        """
        showIndiNewPowerDevice writes info to message window

        :return: true for test purpose
        """

        if deviceName == self.app.power.name:
            self.app.message.emit(f'INDI power device [{deviceName}] found', 0)
        else:
            self.app.message.emit(f'INDI power device snoops -> [{deviceName}]', 0)
        return True

    def showIndiRemovePowerDevice(self, deviceName):
        """
        showIndiRemovePowerDevice writes info to message window

        :return: true for test purpose
        """

        self.app.message.emit(f'INDI power device [{deviceName}] removed', 0)
        return True

    def showPowerDeviceConnected(self):
        """
        showPowerDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.deviceStat['power'] = True
        self.ui.powerDevice.setStyleSheet(self.BACK_GREEN)
        return True

    def showPowerDeviceDisconnected(self):
        """
        showPowerDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :return: true for test purpose
        """

        self.deviceStat['power'] = False
        self.ui.powerDevice.setStyleSheet(self.BACK_NORM)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import time
import pkg_resources
# external packages
import PyQt5.QtCore
import PyQt5.QtWidgets
import PyQt5.uic
# local import
from mw4.gui import widget
from mw4.gui.widgets import message_ui
from indibase import qtIndiBase


class MessageWindow(widget.MWidget):
    """
    the message window class handles

    """

    __all__ = ['MessageWindow',
               ]
    version = '0.2'
    logger = logging.getLogger(__name__)

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.ui = message_ui.Ui_MessageDialog()
        self.ui.setupUi(self)
        self.initUI()

        self.messColor = [self.COLOR_ASTRO,
                          self.COLOR_WHITE,
                          self.COLOR_YELLOW,
                          self.COLOR_RED,
                          ]
        self.messFont = [PyQt5.QtGui.QFont.Normal,
                         PyQt5.QtGui.QFont.Bold,
                         PyQt5.QtGui.QFont.Normal,
                         PyQt5.QtGui.QFont.Normal,
                         ]

        self.initConfig()
        self.showWindow()

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        if 'messageW' not in self.app.config:
            self.app.config['messageW'] = {}
        config = self.app.config['messageW']
        x = config.get('winPosX', 100)
        y = config.get('winPosY', 100)
        if x > self.screenSizeX:
            x = 0
        if y > self.screenSizeY:
            y = 0
        self.move(x, y)
        height = config.get('height', 600)
        self.resize(800, height)

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        if 'messageW' not in self.app.config:
            self.app.config['messageW'] = {}
        config = self.app.config['messageW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()

        return True

    def closeEvent(self, closeEvent):
        self.storeConfig()

        # gui signals
        self.ui.clear.clicked.disconnect(self.clearWindow)
        self.app.message.disconnect(self.writeMessage)

        super().closeEvent(closeEvent)

    def showWindow(self):
        self.show()

        # write basic data to message window
        profile = self.app.config.get('profileName', '-')
        verIB = pkg_resources.get_distribution('indibase').version
        verMC = pkg_resources.get_distribution('mountcontrol').version
        verMW = pkg_resources.get_distribution('mw4').version

        self.writeMessage('MountWizzard4 started', 1)
        self.writeMessage(f'build version: [{verMW}]', 1)
        self.writeMessage(f'mountcontrol version: [{verMC}]', 1)
        self.writeMessage(f'indibase version: [{verIB}]', 1)
        self.writeMessage(f'Workdir is: [{self.app.mwGlob["workDir"]}]', 1)
        self.writeMessage(f'Profile [{profile}] loaded', 0)

        # gui signals
        self.ui.clear.clicked.connect(self.clearWindow)
        self.app.message.connect(self.writeMessage)

    def clearWindow(self):
        """
        clearWindow resets the window and shows empty text.

        :return: true for test purpose
        """

        self.ui.message.clear()
        return True

    def writeMessage(self, message, mType=0):
        """
        writeMessage takes singles with message and writes them to the text browser window.
        types:
            0: normal text
            1: highlighted text
            2: warning text
            3: error text

        :param message: message text
        :param mType: message type
        :return: true for test purpose
        """

        if mType < 0:
            return False
        if mType > len(self.messColor):
            return False
        prefix = time.strftime('%H:%M:%S - ', time.localtime())
        message = prefix + message
        self.logger.info('Message window: [{0}]'.format(message))
        self.ui.message.setTextColor(self.messColor[mType])
        self.ui.message.setFontWeight(self.messFont[mType])
        self.ui.message.insertPlainText(message + '\n')
        self.ui.message.moveCursor(PyQt5.QtGui.QTextCursor.End)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os
import json
import random
# external packages
import numpy as np
import skyfield.api
# local imports
from mw4.base import transform

__all__ = ['HaDecToAltAz',
           ]

version = '0.100.0'


def HaDecToAltAz(ha, dec, lat):
    """
    HaDecToAltAz is derived from http://www.stargazing.net/kepler/altaz.html

    :param ha: hour angle in [h]
    :param dec: declination in [deg]
    :param lat: latitude of observer
    :return: altitude and azimuth
    """

    ha = (ha * 360 / 24 + 360.0) % 360.0
    dec = np.radians(dec)
    ha = np.radians(ha)
    lat = np.radians(lat)
    alt = np.arcsin(np.sin(dec) * np.sin(lat) + np.cos(dec) * np.cos(lat) * np.cos(ha))
    value = (np.sin(dec) - np.sin(alt) * np.sin(lat)) / (np.cos(alt) * np.cos(lat))
    value = np.clip(value, -1.0, 1.0)
    A = np.arccos(value)
    if np.sin(ha) >= 0.0:
        az = 2 * np.pi - A
    else:
        az = A
    az = np.degrees(az)
    alt = np.degrees(alt)
    return alt, az


class DataPoint(object):
    """
    The class Data inherits all information and handling of modeldata data and other
    attributes. this includes horizon data, model points data and their persistence

        >>> data = DataPoint(
        >>>                  app=None,
        >>>                  mwGlob=mwglob,
        >>>                  )
    """

    __all__ = ['DataPoint',
               'genGreaterCircle',
               'genGrid',
               'genInitial',
               'loadBuildP',
               'saveBuildP',
               'clearPoints'
               'deleteBelowHorizon',
               'sort',
               'loadHorizonP',
               'saveHorizonP',
               'clearHorizonP',
               'generateCelestialEquator',
               'generateDSOPath',
               'genAlign',
               'hip',
               ]
    version = '0.100.0'
    logger = logging.getLogger(__name__)

    # data for generating greater circles, dec and step only for east, west is reversed
    DEC = {'min': [-15, 0, 15, 30, 45, 60, 75],
           'norm': [-15, 0, 15, 30, 45, 60, 75],
           'med': [-15, -5, 5, 15, 25, 40, 55, 70, 85],
           'max': [-15, -5, 5, 15, 25, 35, 45, 55, 65, 75, 85],
           }
    STEP = {'min': [15, -15, 15, -15, 15, -30, 30],
            'norm': [10, -10, 10, -10, 10, -20, 20],
            'med': [10, -10, 10, -10, 10, -10, 10, -30, 30],
            'max': [10, -10, 10, -10, 10, -10, 10, -10, 10, -30, 30],
            }
    START = {'min': [-120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5],
             'norm': [-120, -5, -120, -5, -120, -5, -120,
                      5, 120, 5, 120, 5, 120, 5],
             'med': [-120, -5, -120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5, 120, 5],
             'max': [-120, -5, -120, -5, -120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5, 120, 5, 120, 5],
             }
    STOP = {'min': [0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120],
            'norm': [0, -120, 0, -120, 0, -120, 0,
                     120, 0, 120, 0, 120, 0, 120],
            'med': [0, -120, 0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120, 0, 120, 0],
            'max': [0, -120, 0, -120, 0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0],
            }

    def __init__(self,
                 app=None,
                 mwGlob=None,
                 ):

        self.mwGlob = mwGlob
        self.app = app
        self._horizonP = [(0, 0), (0, 360)]
        self._buildP = list()

    @property
    def buildP(self):
        return self._buildP

    @buildP.setter
    def buildP(self, value):
        if not isinstance(value, list):
            self._buildP = list()
            return
        if not all([isinstance(x, tuple) for x in value]):
            self.logger.error('malformed value: {0}'.format(value))
            self._buildP = list()
            return
        self._buildP = value

    def addBuildP(self, value=None, position=None):
        """
        addBuildP extends the list of modeldata points. the new point could be added at the end
        of the list (default) or in any location in the list.

        :param value: value to be inserted
        :param position: position in list
        :return:
        """

        if value is None:
            return False
        if not isinstance(value, tuple):
            self.logger.error('malformed value: {0}'.format(value))
            return False
        if len(value) != 2:
            self.logger.error('malformed value: {0}'.format(value))
            return False
        if position is None:
            position = len(self._buildP)
        if not isinstance(position, (int, float)):
            self.logger.error('malformed position: {0}'.format(position))
            return False
        if self.app.mount.sett.horizonLimitHigh is not None:
            high = self.app.mount.sett.horizonLimitHigh
        else:
            high = 90
        if self.app.mount.sett.horizonLimitLow is not None:
            low = self.app.mount.sett.horizonLimitLow
        else:
            low = 0
        if value[0] > high:
            return False
        if value[0] < low:
            return False
        position = int(position)
        position = min(len(self._buildP), position)
        position = max(0, position)
        self._buildP.insert(position, value)
        return True

    def delBuildP(self, position):
        """
        delBuildP deletes one point from the modeldata points list at the given index.

        :param position:
        :return:
        """

        if not isinstance(position, (int, float)):
            self.logger.error('malformed position: {0}'.format(position))
            return False
        position = int(position)
        if position < 0 or position > len(self._buildP) - 1:
            self.logger.error('invalid position: {0}'.format(position))
            return False
        self._buildP.pop(position)
        return True

    def clearBuildP(self):
        self._buildP = list()

    def checkHorizonBoundaries(self):
        if self._horizonP[0] != (0, 0):
            self._horizonP.insert(0, (0, 0))
        horMax = len(self._horizonP)
        if self._horizonP[horMax - 1] != (0, 360):
            self._horizonP.insert(horMax, (0, 360))

    @property
    def horizonP(self):
        self._horizonP = sorted(self._horizonP, key=lambda x: x[1])
        self.checkHorizonBoundaries()
        return self._horizonP

    @horizonP.setter
    def horizonP(self, value):
        if not isinstance(value, list):
            self.clearHorizonP()
            return
        if not all([isinstance(x, tuple) for x in value]):
            self.logger.error('malformed value: {0}'.format(value))
            self.clearHorizonP()
            return
        self._horizonP = value
        self.checkHorizonBoundaries()

    @staticmethod
    def checkFormat(value):
        if not isinstance(value, list):
            return False
        if not all([isinstance(x, list) for x in value]):
            return False
        if not all([len(x) == 2 for x in value]):
            return False
        return True

    def addHorizonP(self, value=None, position=None):
        """
        addHorizonP extends the list of modeldata points. the new point could be added at the
        end of the list (default) or in any location in the list.

        :param value:
        :param position:
        :return:
        """

        if value is None:
            return False
        if not isinstance(value, tuple):
            self.logger.error('malformed value: {0}'.format(value))
            return False
        if len(value) != 2:
            self.logger.error('malformed value: {0}'.format(value))
            return False
        if position is None:
            position = len(self._horizonP)
        if not isinstance(position, (int, float)):
            self.logger.error('malformed position: {0}'.format(position))
            return False
        position = int(position)
        position = min(len(self._horizonP), position)
        position = max(0, position)
        self._horizonP.insert(position, value)
        return True

    def delHorizonP(self, position):
        """
        delHorizonP deletes one point from the modeldata points list at the given index.

        :param position:
        :return:
        """

        if not isinstance(position, (int, float)):
            self.logger.error('malformed position: {0}'.format(position))
            return False
        position = int(position)
        if position < 0 or position > len(self._horizonP) - 1:
            self.logger.error('invalid position: {0}'.format(position))
            return False
        if self._horizonP[position] == (0, 0):
            return False
        if self._horizonP[position] == (0, 360):
            return False
        self._horizonP.pop(position)
        return True

    def clearHorizonP(self):
        self._horizonP = [(0, 0), (0, 360)]

    def isAboveHorizon(self, point):
        """
        isAboveHorizon calculates for a given point the relationship to the actual horizon
        and determines if this point is above the horizon line. for that there will be a
        linear interpolation for the horizon line points.

        :param point:
        :return:
        """

        if point[1] > 360:
            point = (point[0], 360)
        if point[1] < 0:
            point = (point[0], 0)
        x = range(0, 361)
        y = np.interp(x,
                      [i[1] for i in self._horizonP],
                      [i[0] for i in self._horizonP],
                      )
        if point[0] > y[int(point[1])]:
            return True
        else:
            return False

    def deleteBelowHorizon(self):
        self._buildP = [x for x in self._buildP if self.isAboveHorizon(x)]

    def sort(self, eastwest=False, highlow=False):
        """

        :param eastwest: flag if to be sorted east - west
        :param highlow:  flag if sorted high low altitude
        :return: true for test purpose
        """

        if eastwest and highlow:
            return False
        if not eastwest and not highlow:
            return False

        east = [x for x in self._buildP if x[1] <= 180]
        west = [x for x in self._buildP if x[1] > 180]

        if eastwest:
            east = sorted(east, key=lambda x: -x[1])
            west = sorted(west, key=lambda x: -x[1])

        if highlow:
            east = sorted(east, key=lambda x: -x[0])
            west = sorted(west, key=lambda x: -x[0])

        self._buildP = east + west
        return True

    def loadBuildP(self, fileName=None):
        """
        loadBuildP loads a modeldata pints file and stores the data in the buildP list.
        necessary conversion are made.

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False
        fileName = self.mwGlob['configDir'] + '/' + fileName + '.bpts'
        if not os.path.isfile(fileName):
            return False

        try:
            with open(fileName, 'r') as handle:
                value = json.load(handle)
        except Exception as e:
            self.logger.error('Cannot load: {0}, error: {1}'.format(fileName, e))
            return False

        suc = self.checkFormat(value)
        if not suc:
            self.clearBuildP()
            return False
        # json makes list out of tuple, was to be reversed
        value = [tuple(x) for x in value]
        self._buildP = value
        return True

    def saveBuildP(self, fileName=None):
        """
        saveBuildP saves the actual modeldata points list in a file in json dump format

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False
        fileName = self.mwGlob['configDir'] + '/' + fileName + '.bpts'
        with open(fileName, 'w') as handle:
            json.dump(self.buildP,
                      handle,
                      indent=4)
        return True

    def loadHorizonP(self, fileName=None):
        """
        loadHorizonP loads a modeldata pints file and stores the data in the buildP list.
        necessary conversion are made.

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False
        fileName = self.mwGlob['configDir'] + '/' + fileName + '.hpts'
        if not os.path.isfile(fileName):
            return False

        try:
            with open(fileName, 'r') as handle:
                value = json.load(handle)
        except Exception as e:
            self.logger.error('Cannot load: {0}, error: {1}'.format(fileName, e))
            return False

        suc = self.checkFormat(value)
        if not suc:
            self.clearHorizonP()
            return False
        # json makes list out of tuple, was to be reversed
        value = [tuple(x) for x in value]
        self._horizonP = value
        return True

    @staticmethod
    def checkBoundaries(points):
        """
        checkBoundaries removes point 0,0 and 0, 360 if present.

        :param points:
        :return: points
        """

        if points[0] == (0, 0):
            del points[0]
        if points[len(points) - 1] == (0, 360):
            del points[-1]
        return points

    def saveHorizonP(self, fileName=None):
        """
        saveHorizonP saves the actual modeldata points list in a file in json dump format

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False
        fileName = self.mwGlob['configDir'] + '/' + fileName + '.hpts'
        points = self.checkBoundaries(self.horizonP)
        with open(fileName, 'w') as handle:
            json.dump(points,
                      handle,
                      indent=4)
        return True

    def genHaDecParams(self, selection):
        """
        genHaDecParams selects the parameters for generating the boundaries for next
        step processing greater circles. the parameters are sorted for different targets
        actually for minimum slew distance between the points. defined is only the east
        side of data, the west side will be mirrored to the east one.

        :param selection: type of model we would like to use
        :return: yield tuple of dec value and step, start and stop for range
        """

        if selection not in self.DEC or selection not in self.STEP:
            return
        eastDec = self.DEC[selection]
        westDec = list(reversed(eastDec))
        decL = eastDec + westDec

        eastStepL = self.STEP[selection]
        westStepL = list(reversed(eastStepL))
        stepL = eastStepL + westStepL
        startL = self.START[selection]
        stopL = self.STOP[selection]

        for dec, step, start, stop in zip(decL, stepL, startL, stopL):
            yield dec, step, start, stop

    def genGreaterCircle(self, selection='norm'):
        """
        genGreaterCircle takes the generated boundaries for the rang routine and
        transforms ha, dec to alt az. reasonable values for the alt az values
        are 5 to 85 degrees.

        :param selection:
        :return: yields alt, az tuples which are above horizon
        """

        self.clearBuildP()
        lat = self.app.mount.obsSite.location.latitude.degrees
        for dec, step, start, stop in self.genHaDecParams(selection):
            for ha in range(start, stop, step):
                alt, az = HaDecToAltAz(ha / 10, dec, lat)
                # only values with above horizon = 0

                if 5 <= alt <= 85 and 2 < az < 358:
                    alt += random.uniform(-2, 2)
                    self.addBuildP((alt, az))
        return True

    @staticmethod
    def genGridGenerator(eastAlt, westAlt, minAz, stepAz, maxAz):
        """
        genGridGenerator generates the point values out of the given ranges of altitude
        and azimuth

        :param eastAlt:
        :param westAlt:
        :param minAz:
        :param stepAz:
        :param maxAz:
        :return:
        """

        for i, alt in enumerate(eastAlt):
            if i % 2:
                for az in range(minAz, 180, stepAz):
                    yield(alt, az)
            else:
                for az in range(180 - minAz, 0, -stepAz):
                    yield(alt, az)
        for i, alt in enumerate(westAlt):
            if i % 2:
                for az in range(180 + minAz, 360, stepAz):
                    yield(alt, az)
            else:
                for az in range(maxAz, 180, -stepAz):
                    yield(alt, az)

    def genGrid(self, minAlt=5, maxAlt=85, numbRows=5, numbCols=6):
        """
        genGrid generates a grid of points and transforms ha, dec to alt az. with given
        limits in alt, the min and max will be used as a hard condition. on az there is
        not given limit, therefore a split over the whole space (omitting the meridian)
        is done. the simplest way to avoid hitting the meridian is to enforce the number
        of cols to be a factor of 2. reasonable values for the grid are 5 to 85 degrees.
        defined is only the east side of data, the west side will be mirrored to the
        east one.
            the number of rows is 2 < x < 8
            the number of columns is 2 < x < 15

        :param minAlt: altitude min
        :param maxAlt: altitude max
        :param numbRows: numbRows
        :param numbCols: numbCols
        :return: yields alt, az tuples which are above horizon
        """

        if not 5 <= minAlt <= 85:
            return False
        if not 5 <= maxAlt <= 85:
            return False
        if not maxAlt > minAlt:
            return False
        if not 1 < numbRows < 9:
            return False
        if not 1 < numbCols < 16:
            return False
        if numbCols % 2:
            return False

        minAlt = int(minAlt)
        maxAlt = int(maxAlt)
        numbCols = int(numbCols)
        numbRows = int(numbRows)

        stepAlt = int((maxAlt - minAlt) / (numbRows - 1))
        eastAlt = list(range(minAlt, maxAlt + 1, stepAlt))
        westAlt = list(reversed(eastAlt))

        stepAz = int(360 / numbCols)
        minAz = int(180 / numbCols)
        maxAz = 360 - minAz

        self.clearBuildP()
        for point in self.genGridGenerator(eastAlt, westAlt, minAz, stepAz, maxAz):
            self.addBuildP(point)
        return True

    def genAlign(self, altBase=30, azBase=10, numberBase=3):
        """
        genAlign generates a number of initial points for the first step of modeling. it
        adjusts the first align point in a matter, that the starting point is the closest
        to az = 0.

        :param altBase:
        :param azBase:
        :param numberBase:
        :return: yields alt, az tuples which are above horizon
        """

        if not 5 <= altBase <= 85:
            return False
        if not 2 < numberBase < 11:
            return False
        if not 0 <= azBase < 360:
            return False

        stepAz = int(360 / numberBase)
        altBase = int(altBase)
        azBase = int(azBase) % stepAz
        numberBase = int(numberBase)

        self.clearBuildP()
        for i in range(0, numberBase):
            az = azBase + i * stepAz
            self.addBuildP((altBase, az % 360))
        return True

    def generateCelestialEquator(self):
        """
        generateCelestialEquator calculates a line for greater circles like a celestial
        equator for showing the paths in the hemisphere window.

        :return: celestial equator
        """

        celestialEquator = list()
        lat = self.app.mount.obsSite.location.latitude.degrees
        for dec in range(-15, 90, 15):
            for ha in range(- 119, 120, 3):
                az, alt = HaDecToAltAz(ha / 10, dec, lat)
                if alt > 0:
                    celestialEquator.append((az, alt))
        return celestialEquator

    def generateDSOPath(self, ra=0, dec=0, timeJD=0, location=None,
                        numberPoints=0, duration=0, timeShift=0):
        """
        generateDSOPath calculates a list of model points along the desired path beginning
        at ra, dec coordinates, which is in time duration hours long and consists of
        numberPoints model points. TimeShift moves the pearl of points to an earlier or
        later point in time.

        :param ra:
        :param dec:
        :param timeJD:
        :param location:
        :param numberPoints:
        :param duration:
        :param timeShift:
        :return: True for test purpose
        """

        if numberPoints < 1:
            return False
        if duration == 0:
            return False
        if location is None:
            return False

        numberPoints = int(numberPoints)

        self.clearBuildP()
        for i in range(0, numberPoints):
            startPoint = ra.hours - i * duration / numberPoints - timeShift
            raCalc = skyfield.api.Angle(hours=startPoint)
            az, alt = transform.J2000ToAltAz(raCalc, dec, timeJD, location)
            if alt.degrees > 0:
                self.addBuildP((alt.degrees, az.degrees % 360))

        return True

    def generateGoldenSpiral(self, numberPoints):
        """
        based on the evaluations and implementation of CR Drost from 17-05-24 found at:
        https://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
        the implementation of an equally distributed points cloud over on half of the
        hemisphere.

        :param numberPoints:
        :return: true for test purpose
        """

        self.clearBuildP()

        indices = np.arange(0, numberPoints, dtype=float) + 0.5
        phi = np.arccos(1 - 2 * indices / numberPoints)
        theta = np.pi * (1 + 5 ** 0.5) * indices

        # do not transfer to xyz coordinates
        # x, y, z = np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)
        altitude = 90 - np.degrees(phi)
        azimuth = np.degrees(theta) % 360

        for alt, az in zip(altitude, azimuth):
            # only adding above horizon
            if alt > 0:
                self.addBuildP((alt, az))
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import os
from pathlib import Path
# external packages
import PyQt5
from astropy.io import fits
# local import


class Tools(object):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    def __init__(self):
        self.ui.renameStart.clicked.connect(self.renameRunGUI)
        self.ui.renameInputSelect.clicked.connect(self.chooseDir)

        self.selectorsDropDowns = {'rename1': self.ui.rename1,
                                   'rename2': self.ui.rename2,
                                   'rename3': self.ui.rename3,
                                   'rename4': self.ui.rename4,
                                   'rename5': self.ui.rename5,
                                   }
        self.fitsHeaderKeywords = {'None': [''],
                                   'Datetime': ['DATE-OBS'],
                                   'Frame': ['FRAME', 'IMAGETYP'],
                                   'Filter': ['FILTER'],
                                   'Binning': ['XBINNING'],
                                   'Exp Time': ['EXPTIME'],
                                   'CCD Temp': ['CCD-TEMP'],
                                   }

        self.setupSelectorGui()

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        defaultDir = self.app.mwGlob['imageDir']
        self.ui.renameDir.setText(config.get('renameDir', defaultDir))
        self.ui.checkIncludeSubdirs.setChecked(config.get('checkIncludeSubdirs', False))
        for name, ui in self.selectorsDropDowns.items():
            ui.setCurrentIndex(config.get(name, 0))

        self.ui.renameProgress.setValue(0)
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['renameDir'] = self.ui.renameDir.text()
        config['checkIncludeSubdirs'] = self.ui.checkIncludeSubdirs.isChecked()
        for name, ui in self.selectorsDropDowns.items():
            config[name] = ui.currentIndex()
        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """

        self.wIcon(self.ui.renameStart, PyQt5.QtWidgets.QStyle.SP_DialogApplyButton)

        return True

    def setupSelectorGui(self):
        """
        setupSelectorGui handles the dropdown lists for all devices possible in
        mountwizzard. therefore we add the necessary entries to populate the list.

        :return: success for test
        """

        for name, selectorUI in self.selectorsDropDowns.items():
            selectorUI.clear()
            selectorUI.setView(PyQt5.QtWidgets.QListView())
            for headerEntry in self.fitsHeaderKeywords:
                selectorUI.addItem(headerEntry)

        return True

    @staticmethod
    def getNumberFiles(pathDir='', search=''):
        """
        getNumberFiles counts the number of files to be valid for the renaming.

        :param pathDir: path to root directory to be scanned
        :param search: search string
        :return: number of files found
        """

        if not pathDir:
            return 0
        if not os.path.isdir(pathDir):
            return 0
        if not search:
            return 0

        number = sum(1 for _ in Path(pathDir).glob(search))
        return number

    @staticmethod
    def convertHeaderEntry(entry='', fitsKey=''):
        """
        convertHeaderEntry takes the fitsHeader entry and reformat it to a reasonable
        string.

        :param entry:
        :param fitsKey:
        :return:
        """

        if not fitsKey:
            return ''
        if not entry:
            return ''

        if fitsKey == 'DATE-OBS':
            chunk = entry.replace(':', '-')
            chunk = chunk.replace('T', '-')
            chunk = chunk.split('.')[0]
        elif fitsKey == 'XBINNING':
            chunk = f'Bin{entry}'
        elif fitsKey == 'CCD-TEMP':
            chunk = f'Temp{entry:03.0f}'
        elif fitsKey == 'FRAME':
            chunk = f'{entry}'
        elif fitsKey == 'FILTER':
            chunk = f'{entry}'
        else:
            chunk = ''

        return chunk

    def processSelectors(self, fitsHeader=None, selection=''):
        """
        processSelectors takes the selection for a fileName chunk and runs through the
        possible list of valid fits header keys. if there is more than one valid fitsKey,
        it automatically selects only the first on for conversion.

        :param fitsHeader:
        :param selection: str entry from the drop down selector
        :return: nameChunk: part of the entry
        """

        if fitsHeader is None:
            return ''
        if not selection:
            return ''

        nameChunk = ''
        fitsKeywords = self.fitsHeaderKeywords[selection]
        for fitsKey in fitsKeywords:
            if fitsKey not in fitsHeader:
                continue
            nameChunk = self.convertHeaderEntry(entry=fitsHeader[fitsKey],
                                                fitsKey=fitsKey)
            break
        return nameChunk

    def renameFile(self, fileName=''):
        """
        renameFile opens the given FITS file and retrieves it's header. if valid it
        runs through selectors of the drop down lists and checks all header keys to
        get the new filename build. afterwards it renames the given file.

        :param fileName: fits file to be renamed
        :return: success
        """

        if not fileName:
            return False

        with fits.open(name=fileName) as fd:
            fitsHeader = fd[0].header

            # object should bi in in any case. if not, it will be set
            if 'OBJECT' in fitsHeader:
                newFilename = fitsHeader['OBJECT'].lower()
            else:
                newFilename = 'unknown'

            for _, selector in self.selectorsDropDowns.items():
                selection = selector.currentText()
                chunk = self.processSelectors(fitsHeader=fitsHeader,
                                              selection=selection
                                              )
                if chunk:
                    newFilename += f'-{chunk}'

            newFilename += '.fits'

            dirName = os.path.dirname(fileName)
            newFilename = f'{dirName}/{newFilename}'
            os.rename(fileName, newFilename)

        return True

    def renameRunGUI(self):
        """
        renameRunGUI retrieves a full list of files to be renamed and renames
        them on by one.

        :return: True for test purpose
        """

        pathDir = self.ui.renameDir.text()
        includeSubdirs = self.ui.checkIncludeSubdirs.isChecked()

        if not os.path.isdir(pathDir):
            self.app.message.emit('No valid input directory given', 2)
            return False

        if includeSubdirs:
            search = '**/*.fit*'
        else:
            search = '*.fit*'

        numberFiles = self.getNumberFiles(pathDir, search=search)
        if not numberFiles:
            self.app.message.emit('No files to rename', 0)
            return False

        for i, fileName in enumerate(Path(pathDir).glob(search)):
            self.ui.renameProgress.setValue(int(100 * (i + 1) / numberFiles))
            PyQt5.QtWidgets.QApplication.processEvents()
            suc = self.renameFile(fileName=fileName)
            if not suc:
                self.app.message.emit(f'{fileName} could not be renamed', 2)

        self.app.message.emit(f'{numberFiles:d} images were renamed', 0)

        return True

    def chooseDir(self):
        """
        chooseDir selects the input directory and sets the default value for the
        output directory as well

        :return: True for test purpose
        """
        folder = self.ui.renameDir.text()
        pathDir, _, _ = self.openDir(self,
                                     'Choose Input Dir',
                                     folder,
                                     )
        if pathDir:
            self.ui.renameDir.setText(pathDir)
            self.ui.renameProgress.setValue(0)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
# external packages
import numpy as np
import astropy._erfa as erfa
# local imports
from mw4.modeldata.alignstars import generateAlignStars


class Hipparcos(object):
    """
    The class Data inherits all information and handling of hipparcos data and other
    attributes. this includes data about the alignment stars defined in generateAlignStars,
    their ra dec coordinates, proper motion, parallax and radial velocity and the
    calculation of data for display and slew commands

        >>> hip = Hipparcos(
        >>>                 app=app
        >>>                 mwGlob=mwglob
        >>>                 )
    """

    __all__ = ['Hipparcos',
               'calculateAlignStarsPositionsAltAz',
               'getAlignStarRaDecFromIndex',
               ]
    version = '0.100.0'
    logger = logging.getLogger(__name__)

    def __init__(self,
                 app=None,
                 mwGlob=None,
                 ):

        self.app = app
        self.mwGlob = mwGlob
        self.lat = app.mount.obsSite.location.latitude.degrees
        self.name = list()
        self.alt = list()
        self.az = list()
        self.alignStars = generateAlignStars()
        self.calculateAlignStarPositionsAltAz()

    def calculateAlignStarPositionsAltAz(self):
        """
        calculateAlignStarPositionsAltAz does calculate the star coordinates from give data
        out of generated star list. calculation routines are from astropy erfa. atco13 does
        the results based on proper motion, parallax and radial velocity and need J2000
        coordinates. because of using the hipparcos catalogue, which is based on J1991,
        25 epoch the pre calculation from J1991,25 to J2000 is done already when generating
        the alignstars file. there is no refraction data taken into account, because we need
        this only for display purpose and for this, the accuracy is more than sufficient.

        :return: lists for alt, az and name of star
        """

        location = self.app.mount.obsSite.location
        if location is None:
            return False
        t = self.app.mount.obsSite.timeJD
        star = list(self.alignStars.values())
        self.name = list(self.alignStars.keys())

        aob, zob, hob, dob, rob, eo = erfa.atco13([x[0] for x in star],
                                                  [x[1] for x in star],
                                                  [x[2] for x in star],
                                                  [x[3] for x in star],
                                                  [x[4] for x in star],
                                                  [x[5] for x in star],
                                                  t.ut1,
                                                  0.0,
                                                  t.dut1,
                                                  location.longitude.radians,
                                                  location.latitude.radians,
                                                  location.elevation.m,
                                                  0.0,
                                                  0.0,
                                                  0.0,
                                                  0.0,
                                                  0.0,
                                                  0.0)
        self.az = aob * 360 / 2 / np.pi
        self.alt = 90.0 - zob * 360 / 2 / np.pi
        return True

    def getAlignStarRaDecFromName(self, name):
        """
        getAlignStarRaDecFromName does calculate the star coordinates from give data
        out of generated star list. calculation routines are from astropy erfa. atco13 does
        the results based on proper motion, parallax and radial velocity and need J2000
        coordinates. because of using the hipparcos catalogue, which is based on J1991,
        25 epoch the pre calculation from J1991,25 to J2000 is done already when generating
        the alignstars file. there is no refraction data taken into account, because we need
        this only for display purpose and for this, the accuracy is more than sufficient.

        the return values are in JNow epoch as the mount only handles this epoch !

        :param name: name of star
        :return: values for ra, dec in hours / degrees in JNow epoch !
        """

        if name not in self.alignStars:
            return None, None
        t = self.app.mount.obsSite.ts.now()
        values = self.alignStars[name]

        ra, dec, eo = erfa.atci13(values[0],
                                  values[1],
                                  values[2],
                                  values[3],
                                  values[4],
                                  values[5],
                                  t.ut1,
                                  0.0,
                                  )
        ra = erfa.anp(ra - eo) * 24 / 2 / np.pi
        dec = dec * 360 / 2 / np.pi

        return ra, dec



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os
# external packages
import PyQt5.QtWidgets
from astropy.io import fits
from astropy import wcs
from astropy.nddata import Cutout2D
from astropy.visualization import AsymmetricPercentileInterval
from astropy.visualization import SqrtStretch
from astropy.visualization import ImageNormalize
import matplotlib.pyplot as plt
import numpy as np
# local import
from mw4.gui import widget
from mw4.gui.widgets import image_ui
from mw4.base import transform


class ImageWindowSignals(PyQt5.QtCore.QObject):
    """
    The CameraSignals class offers a list of signals to be used and instantiated by
    the Mount class to get signals for triggers for finished tasks to
    enable a gui to update their values transferred to the caller back.

    This has to be done in a separate class as the signals have to be subclassed from
    QObject and the Mount class itself is subclassed from object
    """

    __all__ = ['ImageWindowSignals']
    version = '0.101'

    showCurrent = PyQt5.QtCore.pyqtSignal()
    showImage = PyQt5.QtCore.pyqtSignal(object)
    solveImage = PyQt5.QtCore.pyqtSignal(object)


class ImageWindow(widget.MWidget):
    """
    the image window class handles fits image loading, stretching, zooming and handles
    the gui interface for display. both wcs and pixel coordinates will be used.

    """

    __all__ = ['ImageWindow',
               ]
    version = '0.101'
    logger = logging.getLogger(__name__)

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.ui = image_ui.Ui_ImageDialog()
        self.ui.setupUi(self)
        self.initUI()
        self.signals = ImageWindowSignals()

        self.imageFileName = ''
        self.imageFileNameOld = ''
        self.imageStack = None
        self.raStack = 0
        self.decStack = 0
        self.angleStack = 0
        self.scaleStack = 0
        self.numberStack = 0
        self.folder = ''

        self.deviceStat = {
            'expose': False,
            'exposeN': False,
            'solve': False,
        }
        self.colorMaps = {'Grey': 'gray',
                          'Cool': 'plasma',
                          'Rainbow': 'rainbow',
                          'Spectral': 'nipy_spectral',
                          }

        self.stretchValues = {'Low X': (99, 99.999),
                              'Low': (90, 99.995),
                              'Mid': (50, 99.99),
                              'High': (20, 99.98),
                              'Super': (10, 99.9),
                              'Super X': (1, 99.8),
                              }

        self.zoomLevel = {'Zoom 1x': 1,
                          'Zoom 2x': 2,
                          'Zoom 4x': 4,
                          'Zoom 8x': 8,
                          'Zoom 16x': 16,
                          }

        self.imageMat = self.embedMatplot(self.ui.image)
        self.imageMat.parentWidget().setStyleSheet(self.BACK_BG)

        self.initConfig()
        self.showWindow()
        self.app.update1s.connect(self.updateWindowsStats)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method. if not entry is already in the
        config dict, it will be created first.
        default values will be set in case of missing parameters.
        screen size will be set as well as the window position. if the window position is
        out of the current screen size (because of copy configs or just because the screen
        resolution was changed) the window will be repositioned so that it will be visible.

        :return: True for test purpose
        """

        if 'imageW' not in self.app.config:
            self.app.config['imageW'] = {}
        config = self.app.config['imageW']
        x = config.get('winPosX', 100)
        y = config.get('winPosY', 100)
        if x > self.screenSizeX:
            x = 0
        if y > self.screenSizeY:
            y = 0
        self.move(x, y)
        height = config.get('height', 600)
        width = config.get('width', 800)
        self.resize(width, height)

        self.setupDropDownGui()
        self.ui.color.setCurrentIndex(config.get('color', 0))
        self.ui.zoom.setCurrentIndex(config.get('zoom', 0))
        self.ui.stretch.setCurrentIndex(config.get('stretch', 0))
        self.imageFileName = config.get('imageFileName', '')
        self.folder = self.app.mwGlob.get('imageDir', '')
        self.ui.checkUseWCS.setChecked(config.get('checkUseWCS', False))
        self.ui.checkStackImages.setChecked(config.get('checkStackImages', False))
        self.ui.checkShowCrosshair.setChecked(config.get('checkShowCrosshair', False))
        self.ui.checkShowGrid.setChecked(config.get('checkShowGrid', True))
        self.ui.checkAutoSolve.setChecked(config.get('checkAutoSolve', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """

        config = self.app.config['imageW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()
        config['width'] = self.width()
        config['color'] = self.ui.color.currentIndex()
        config['zoom'] = self.ui.zoom.currentIndex()
        config['stretch'] = self.ui.stretch.currentIndex()
        config['imageFileName'] = self.imageFileName
        config['checkUseWCS'] = self.ui.checkUseWCS.isChecked()
        config['checkStackImages'] = self.ui.checkStackImages.isChecked()
        config['checkShowCrosshair'] = self.ui.checkShowCrosshair.isChecked()
        config['checkShowGrid'] = self.ui.checkShowGrid.isChecked()
        config['checkAutoSolve'] = self.ui.checkAutoSolve.isChecked()

        return True

    def showWindow(self):
        """
        showWindow prepares all data for showing the window, plots the image and show it.
        afterwards all necessary signal / slot connections will be established.

        :return: true for test purpose
        """

        self.showImage(self.imageFileName)
        self.show()

        # gui signals
        self.ui.load.clicked.connect(self.selectImage)
        self.ui.color.currentIndexChanged.connect(self.showCurrent)
        self.ui.stretch.currentIndexChanged.connect(self.showCurrent)
        self.ui.zoom.currentIndexChanged.connect(self.showCurrent)
        self.ui.checkUseWCS.clicked.connect(self.showCurrent)
        self.ui.checkShowGrid.clicked.connect(self.showCurrent)
        self.ui.checkShowCrosshair.clicked.connect(self.showCurrent)
        self.ui.solve.clicked.connect(self.solveCurrent)
        self.ui.expose.clicked.connect(self.exposeImage)
        self.ui.exposeN.clicked.connect(self.exposeImageN)
        self.ui.abortImage.clicked.connect(self.abortImage)
        self.ui.abortSolve.clicked.connect(self.abortSolve)
        self.signals.showCurrent.connect(self.showCurrent)
        self.signals.showImage.connect(self.showImage)
        self.signals.solveImage.connect(self.solveImage)

        return True

    def closeEvent(self, closeEvent):
        """
        closeEvent overlays the window close event of qt. first it stores all persistent
        data for the windows and its functions, than removes al signal / slot connections
        removes the matplotlib embedding and finally calls the parent calls for handling
        the framework close event.

        :param closeEvent:
        :return: True for test purpose
        """

        self.storeConfig()

        # gui signals
        self.ui.load.clicked.disconnect(self.selectImage)
        self.ui.color.currentIndexChanged.disconnect(self.showCurrent)
        self.ui.stretch.currentIndexChanged.disconnect(self.showCurrent)
        self.ui.zoom.currentIndexChanged.disconnect(self.showCurrent)
        self.ui.checkUseWCS.clicked.disconnect(self.showCurrent)
        self.ui.checkShowGrid.clicked.disconnect(self.showCurrent)
        self.ui.checkShowCrosshair.clicked.disconnect(self.showCurrent)
        self.ui.solve.clicked.disconnect(self.solveCurrent)
        self.ui.expose.clicked.disconnect(self.exposeImage)
        self.ui.exposeN.clicked.disconnect(self.exposeImageN)
        self.ui.abortImage.clicked.disconnect(self.abortImage)
        self.ui.abortSolve.clicked.disconnect(self.abortSolve)
        self.signals.showCurrent.disconnect(self.showCurrent)
        self.signals.showImage.disconnect(self.showImage)
        self.signals.solveImage.disconnect(self.solveImage)

        plt.close(self.imageMat.figure)
        super().closeEvent(closeEvent)

    def setupDropDownGui(self):
        """
        setupDropDownGui handles the population of list for image handling.

        :return: success for test
        """

        self.ui.color.clear()
        self.ui.color.setView(PyQt5.QtWidgets.QListView())
        for text in self.colorMaps:
            self.ui.color.addItem(text)

        self.ui.zoom.clear()
        self.ui.zoom.setView(PyQt5.QtWidgets.QListView())
        for text in self.zoomLevel:
            self.ui.zoom.addItem(text)

        self.ui.stretch.clear()
        self.ui.stretch.setView(PyQt5.QtWidgets.QListView())
        for text in self.stretchValues:
            self.ui.stretch.addItem(text)

        return True

    def updateWindowsStats(self):
        """
        updateWindowsStats changes dynamically the enable and disable of user gui elements
        depending of the actual state of processing

        :return: true for test purpose

        """

        if self.deviceStat['expose']:
            self.ui.exposeN.setEnabled(False)
            self.ui.load.setEnabled(False)
            self.ui.abortImage.setEnabled(True)

        elif self.deviceStat['exposeN']:
            self.ui.expose.setEnabled(False)
            self.ui.load.setEnabled(False)
            self.ui.abortImage.setEnabled(True)
        else:
            self.ui.solve.setEnabled(True)
            self.ui.expose.setEnabled(True)
            self.ui.exposeN.setEnabled(True)
            self.ui.load.setEnabled(True)
            self.ui.abortImage.setEnabled(False)

        if self.deviceStat['solve']:
            self.ui.abortSolve.setEnabled(True)
        else:
            self.ui.abortSolve.setEnabled(False)

        if not self.app.mainW.deviceStat['imaging']:
            self.ui.expose.setEnabled(False)
            self.ui.exposeN.setEnabled(False)

        if not self.app.mainW.deviceStat['astrometry']:
            self.ui.solve.setEnabled(False)

        if self.deviceStat['expose']:
            self.changeStyleDynamic(self.ui.expose, 'running', 'true')

        elif self.deviceStat['exposeN']:
            self.changeStyleDynamic(self.ui.exposeN, 'running', 'true')

        else:
            self.changeStyleDynamic(self.ui.expose, 'running', 'false')
            self.changeStyleDynamic(self.ui.exposeN, 'running', 'false')

        if self.deviceStat['solve']:
            self.changeStyleDynamic(self.ui.solve, 'running', 'true')
        else:
            self.changeStyleDynamic(self.ui.solve, 'running', 'false')

        return True

    def selectImage(self):
        """
        selectImage does a dialog to choose a FITS file for viewing. The file will not
        be loaded, just the full filepath will be stored. if succeeding, the signal for
        displaying the image will be emitted.

        :return: success
        """

        loadFilePath, name, ext = self.openFile(self,
                                                'Select image file',
                                                self.folder,
                                                'FITS files (*.fit*)',
                                                enableDir=True,
                                                )
        if not name:
            return False

        self.ui.imageFileName.setText(name)
        self.imageFileName = loadFilePath
        self.app.message.emit(f'Image [{name}] selected', 0)
        self.ui.checkUseWCS.setChecked(False)
        self.folder = os.path.dirname(loadFilePath)
        self.signals.showImage.emit(self.imageFileName)

        return True

    def writeHeaderToGUI(self, header=None):
        """
        writeHeaderToGUI tries to read relevant values from FITS header and possible
        replace values and write them to the imageW gui

        :param header: header of fits file
        :return: hasCelestial, hasDistortion
        """

        name = header.get('OBJECT', '').upper()
        self.ui.object.setText(f'{name}')

        ra = header.get('OBJCTRA', 0)
        ra = transform.convertToAngle(ra, isHours=True)
        dec = header.get('OBJCTDEC', 0)
        dec = transform.convertToAngle(dec, isHours=False)
        self.ui.ra.setText(f'{transform.convertToHMS(ra)}')
        self.ui.dec.setText(f'{transform.convertToDMS(dec)}')

        scale = header.get('SCALE', 0)
        rotation = header.get('ANGLE', 0)
        self.ui.scale.setText(f'{scale:5.3f}')
        self.ui.rotation.setText(f'{rotation:6.3f}')

        ccdTemp = header.get('CCD-TEMP', 0)
        self.ui.ccdTemp.setText(f'{ccdTemp:4.1f}')

        expTime1 = header.get('EXPOSURE', 0)
        expTime2 = header.get('EXPTIME', 0)
        expTime = max(expTime1, expTime2)
        self.ui.expTime.setText(f'{expTime:5.1f}')

        filterCCD = header.get('FILTER', 0)
        self.ui.filter.setText(f'{filterCCD}')

        binX = header.get('XBINNING', 0)
        binY = header.get('YBINNING', 0)
        self.ui.binX.setText(f'{binX:1.0f}')
        self.ui.binY.setText(f'{binY:1.0f}')

        sqm = max(header.get('SQM', 0),
                  header.get('SKY-QLTY', 0),
                  header.get('MPSAS', 0),
                  )
        self.ui.sqm.setText(f'{sqm:5.2f}')

        flipped = header.get('FLIPPED', False)
        self.ui.isFlipped.setEnabled(flipped)

        # check if distortion is in header
        if 'CTYPE1' in header:
            wcsObject = wcs.WCS(header, relax=True)
            hasCelestial = wcsObject.has_celestial
            hasDistortion = wcsObject.has_distortion
        else:
            wcsObject = None
            hasCelestial = False
            hasDistortion = False

        self.ui.hasDistortion.setEnabled(hasDistortion)
        self.ui.hasWCS.setEnabled(hasCelestial)

        return hasDistortion, wcsObject

    def zoomImage(self, image=None, wcsObject=None):
        """
        zoomImage cutouts a portion of the original image to zoom in the image itself.
        it returns a copy of the image with an updated wcs content. we have to be careful
        about the use of Cutout2D, because they are mixing x and y coordinates. so position
        is in (x, y), but size ind in (y, x)

        :param image:
        :param wcsObject:
        :return:
        """

        if image is None:
            return None

        sizeY, sizeX = image.shape
        factor = self.zoomLevel[self.ui.zoom.currentText()]
        position = (int(sizeX / 2), int(sizeY / 2))
        size = (int(sizeY / factor), int(sizeX / factor))

        cutout = Cutout2D(image,
                          position=position,
                          size=size,
                          wcs=wcsObject,
                          copy=True,
                          )

        return cutout.data

    def stretchImage(self, image=None):
        """
        stretchImage take the actual image and calculated norm based on the min, max
        derived from interval which is calculated with AsymmetricPercentileInterval.

        :param image: image
        :return: norm for plot
        """

        if image is None:
            return None

        values = self.stretchValues[self.ui.stretch.currentText()]
        interval = AsymmetricPercentileInterval(*values)
        vmin, vmax = interval.get_limits(image)

        norm = ImageNormalize(image,
                              vmin=vmin,
                              vmax=vmax,
                              stretch=SqrtStretch(),
                              )

        return norm

    def colorImage(self):
        """
        colorImage take the index from gui and generates the colormap for image show
        command from matplotlib

        :return: color map
        """

        colorMap = self.colorMaps[self.ui.color.currentText()]

        return colorMap

    def setupDistorted(self, figure=None, wcsObject=None):
        """
        setupDistorted tries to setup all necessary context for displaying the image with
        wcs distorted coordinates.
        still plenty of work to be done, because very often the labels are not shown

        :param figure:
        :param wcsObject:
        :return: axes object to plot onto
        """

        if figure is None:
            return False
        if wcsObject is None:
            return False

        figure.clf()
        figure.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9)

        axe = figure.add_subplot(1, 1, 1, projection=wcsObject, facecolor=None)
        axe.coords.frame.set_color(self.M_BLUE)

        axe0 = axe.coords[0]
        axe1 = axe.coords[1]
        if self.ui.checkShowGrid.isChecked():
            axe0.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)
            axe1.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)
        axe0.tick_params(colors=self.M_BLUE, labelsize=12)
        axe1.tick_params(colors=self.M_BLUE, labelsize=12)
        axe0.set_axislabel('Right Ascension',
                           color=self.M_BLUE,
                           fontsize=12,
                           fontweight='bold')
        axe1.set_axislabel('Declination',
                           color=self.M_BLUE,
                           fontsize=12,
                           fontweight='bold')
        axe0.set_ticks(number=10)
        axe1.set_ticks(number=10)
        axe0.set_ticks_position('lr')
        axe0.set_ticklabel_position('lr')
        axe0.set_axislabel_position('lr')
        axe1.set_ticks_position('tb')
        axe1.set_ticklabel_position('tb')
        axe1.set_axislabel_position('tb')

        return axe

    def setupNormal(self, figure=None, header=None):
        """
        setupNormal build the image widget to show it with pixels as axes. the center of
        the image will have coordinates 0,0.

        :param figure:
        :param header:
        :return: axes object to plot onto
        """

        if figure is None:
            return False
        if header is None:
            return False

        figure.clf()
        # figure.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.95)

        axe = figure.add_subplot(1, 1, 1, facecolor=None)

        axe.tick_params(axis='x', which='major', colors=self.M_BLUE, labelsize=12)
        axe.tick_params(axis='y', which='major', colors=self.M_BLUE, labelsize=12)

        factor = self.zoomLevel[self.ui.zoom.currentText()]
        sizeX = header.get('NAXIS1', 1) / factor
        sizeY = header.get('NAXIS2', 1) / factor
        midX = int(sizeX / 2)
        midY = int(sizeY / 2)
        number = 10

        valueX, _ = np.linspace(-midX, midX, num=number, retstep=True)
        textX = list((str(int(x)) for x in valueX))
        ticksX = list((x + midX for x in valueX))
        axe.set_xticklabels(textX)
        axe.set_xticks(ticksX)

        valueY, _ = np.linspace(-midY, midY, num=number, retstep=True)
        textY = list((str(int(x)) for x in valueY))
        ticksY = list((x + midY for x in valueY))
        axe.set_yticklabels(textY)
        axe.set_yticks(ticksY)

        if self.ui.checkShowCrosshair.isChecked():
            axe.axvline(midX, color=self.M_RED)
            axe.axhline(midY, color=self.M_RED)
        if self.ui.checkShowGrid.isChecked():
            axe.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)

        axe.set_xlabel(xlabel='Pixel', color=self.M_BLUE, fontsize=12, fontweight='bold')
        axe.set_ylabel(ylabel='Pixel', color=self.M_BLUE, fontsize=12, fontweight='bold')
        return axe

    def stackImages(self, imageData=None, header=None):
        """

        :param imageData:
        :param header: is only used, when stacking with alignment
        :return:
        """

        # if first image, we just store the data as reference frame
        if self.imageStack is None:
            self.imageStack = imageData
            self.numberStack = 1
            return imageData

        # now we are going to stack the results
        self.numberStack += 1
        self.imageStack = np.add(self.imageStack, imageData)
        return self.imageStack / self.numberStack

    def showImage(self, imagePath=''):
        """
        showImage shows the fits image. therefore it calculates color map, stretch,
        zoom and other topics.

        :param imagePath:
        :return: success
        """

        if not imagePath:
            return False
        if not os.path.isfile(imagePath):
            return False

        self.imageFileName = imagePath
        full, short, ext = self.extractNames([imagePath])
        self.ui.imageFileName.setText(short)

        with fits.open(imagePath, mode='update') as fitsHandle:
            imageData = fitsHandle[0].data
            header = fitsHandle[0].header

            # correct faulty headers, because some imaging programs did not
            # interpret the Keywords in the right manner (SGPro)

            if header.get('CTYPE1', '').endswith('DEF'):
                header['CTYPE1'] = header['CTYPE1'].replace('DEF', 'TAN')
            if header.get('CTYPE2', '').endswith('DEF'):
                header['CTYPE2'] = header['CTYPE2'].replace('DEF', 'TAN')

        # check the data content and capabilities
        hasDistortion, wcsObject = self.writeHeaderToGUI(header=header)

        if self.ui.checkStackImages.isChecked():
            imageData = self.stackImages(imageData=imageData,
                                         header=header)

        # process the image for viewing
        imageData = self.zoomImage(image=imageData, wcsObject=wcsObject)
        norm = self.stretchImage(image=imageData)
        colorMap = self.colorImage()

        # check which type of presentation we would like to have
        useWCS = self.ui.checkUseWCS.isChecked()

        if hasDistortion and useWCS:
            axe = self.setupDistorted(figure=self.imageMat.figure, wcsObject=wcsObject)
        else:
            axe = self.setupNormal(figure=self.imageMat.figure, header=header)

        # finally show it
        axe.imshow(imageData, norm=norm, cmap=colorMap, origin='lower')
        axe.figure.canvas.draw()

        return True

    def showCurrent(self):
        """

        :return: true for test purpose
        """
        self.showImage(self.imageFileName)
        return True

    def exposeRaw(self):
        """
        exposeImage gathers all necessary parameters and starts exposing

        :return: True for test purpose
        """

        expTime = self.app.mainW.ui.expTime.value()
        binning = self.app.mainW.ui.binning.value()
        subFrame = self.app.mainW.ui.subFrame.value()
        fastReadout = self.app.mainW.ui.checkFastDownload.isChecked()

        time = self.app.mount.obsSite.timeJD.utc_strftime('%Y-%m-%d-%H-%M-%S')
        fileName = time + '-exposure.fits'
        imagePath = self.app.mwGlob['imageDir'] + '/' + fileName

        self.imageFileNameOld = self.imageFileName

        self.app.imaging.expose(imagePath=imagePath,
                                expTime=expTime,
                                binning=binning,
                                subFrame=subFrame,
                                fastReadout=fastReadout,
                                )

        self.app.message.emit(f'Exposing: [{os.path.basename(imagePath)}]', 0)
        text = f'Duration: {expTime:3.0f}s  Bin: {binning:1.0f}  Sub: {subFrame:3.0f}%'
        self.app.message.emit(f'     {text}', 0)

        return True

    def exposeImageDone(self, imagePath=''):
        """
        exposeImageDone is the partner method to exposeImage. it resets the gui elements
        to it's default state and disconnects the signal for the callback. finally when
        all elements are done it emits the showImage signal.

        :param imagePath:
        :return: True for test purpose
        """

        self.deviceStat['expose'] = False
        self.app.imaging.signals.saved.disconnect(self.exposeImageDone)
        self.app.message.emit(f'Exposed: [{os.path.basename(imagePath)}]', 0)

        if self.ui.checkAutoSolve.isChecked():
            self.signals.solveImage.emit(imagePath)
        else:
            self.signals.showImage.emit(imagePath)

        return True

    def exposeImage(self):
        """
        exposeImage disables all gui elements which could interfere when having a running
        exposure. it connects the callback for downloaded image and presets all necessary
        parameters for imaging

        :return: success
        """

        self.deviceStat['expose'] = True
        self.app.imaging.signals.saved.connect(self.exposeImageDone)
        self.exposeRaw()

        return True

    def exposeImageNDone(self, imagePath=''):
        """
        exposeImageNDone is the partner method to exposeImage. it resets the gui elements
        to it's default state and disconnects the signal for the callback. finally when
        all elements are done it emits the showImage signal.

        :param imagePath:
        :return: True for test purpose
        """

        self.app.message.emit(f'Exposed: [{os.path.basename(imagePath)}]', 0)

        if self.ui.checkAutoSolve.isChecked():
            self.signals.solveImage.emit(imagePath)
        else:
            self.signals.showImage.emit(imagePath)

        self.exposeRaw()

        return True

    def exposeImageN(self):
        """
        exposeImageN disables all gui elements which could interfere when having a running
        exposure. it connects the callback for downloaded image and presets all necessary
        parameters for imaging

        :return: success
        """

        self.imageStack = None
        self.deviceStat['exposeN'] = True
        self.app.imaging.signals.saved.connect(self.exposeImageNDone)
        self.exposeRaw()

        return True

    def abortImage(self):
        """
        abortImage stops the exposure and resets the gui and the callback signals to default
        values

        :return: True for test purpose
        """

        self.app.imaging.abort()

        # for disconnection we have to split which slots were connected to disable the
        # right ones
        if self.deviceStat['expose']:
            self.app.imaging.signals.saved.disconnect(self.exposeImageDone)
        if self.deviceStat['exposeN']:
            self.app.imaging.signals.saved.disconnect(self.exposeImageNDone)

        # last image file was nor stored, so getting last valid it back
        self.imageFileName = self.imageFileNameOld
        self.deviceStat['expose'] = False
        self.deviceStat['exposeN'] = False
        self.app.message.emit('Exposing aborted', 2)

        return True

    def solveDone(self, result):
        """
        solveDone is the partner method for solveImage. it enables the gui elements back
        removes the signal / slot connection for receiving solving results, checks the
        solving result itself and emits messages about the result. if solving succeeded,
        solveDone will redraw the image in the image window.

        :param result: result (named tuple)
        :return: success
        """

        self.deviceStat['solve'] = False
        self.app.astrometry.signals.done.disconnect(self.solveDone)

        if not result.success:
            self.app.message.emit('Solving error', 2)
            return False

        rData = result.solve
        if not isinstance(rData, tuple):
            return False
        text = f'Ra: {transform.convertToHMS(rData.raJ2000)} '
        text += f'({rData.raJ2000.hours:4.3f}), '
        text += f'Dec: {transform.convertToDMS(rData.decJ2000)} '
        text += f'({rData.decJ2000.degrees:4.3f}), '
        text += f'Error: {rData.error:5.1f}, Angle: {rData.angle:3.0f}, '
        text += f'Scale: {rData.scale:4.3f}'
        if result.success:
            self.app.message.emit(f'Solved: [{os.path.basename(result.solve.path)}]', 0)
            self.app.message.emit(f'     {text}', 0)
        else:
            self.app.message.emit('Solving error', 2)

        isStack = self.ui.checkStackImages.isChecked()
        isAutoSolve = self.ui.checkAutoSolve.isChecked()
        if not isStack or isAutoSolve:
            self.signals.showImage.emit(result.solve.path)

        return True

    def solveImage(self, imagePath=''):
        """
        solveImage calls astrometry for solving th actual image in a threading manner.
        as result the gui will be active while the solving process takes part a
        background task. if the check update fits is selected the solution will be
        stored in the image header as well.
        solveImage will disable gui elements which might interfere when doing solve
        in background and sets the signal / slot connection for receiving the signal
        for finishing. this is linked to a second method solveDone, which is basically
        the partner method for handling this async behaviour of the gui.
        finally it emit a message about the start of solving

        :param imagePath:
        :return:
        """

        if not imagePath:
            return False
        if not os.path.isfile(imagePath):
            return False

        updateFits = True
        solveTimeout = self.app.mainW.ui.solveTimeout.value()
        searchRadius = self.app.mainW.ui.searchRadius.value()
        app = self.app.mainW.ui.astrometryDevice.currentText()
        self.app.astrometry.solveThreading(app=app,
                                           fitsPath=imagePath,
                                           radius=searchRadius,
                                           timeout=solveTimeout,
                                           updateFits=updateFits,
                                           )
        self.deviceStat['solve'] = True
        self.app.astrometry.signals.done.connect(self.solveDone)
        self.app.message.emit(f'Solving: [{os.path.basename(imagePath)}]', 0)

        return True

    def solveCurrent(self):
        """

        :return: true for test purpose
        """

        self.signals.solveImage.emit(self.imageFileName)
        return True

    def abortSolve(self):
        """
        abortSolve stops the exposure and resets the gui and the callback signals to default
        values

        :return: success
        """

        suc = self.app.astrometry.abort()

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.7.4
#
# Michael Würtenberger
# (c) 2019
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os
import sys
import platform
import socket
import datetime
import warnings
import traceback
from io import BytesIO
# external packages
import matplotlib
matplotlib.use('Qt5Agg')
import PyQt5.QtCore
import PyQt5.QtWidgets
# local import
from mw4 import mainApp
from mw4.gui import splash


class MyApp(PyQt5.QtWidgets.QApplication):
    """
    MyApp implements a custom notify handler to log errors, when C++ classes and python
    wrapper in PyQt5 environment mismatch. mostly this relates to the situation when a
    C++ object is already deleted, but the python wrapper still exists. so far I know
    that's the only chance to log this issues.

    in addition it writes mouse pressed and key pressed events in debug level to log
    including event and object name to be analyse the input methods.
    """

    def notify(self, obj, event):
        try:
            returnValue = PyQt5.QtWidgets.QApplication.notify(self, obj, event)
        except Exception as e:
            logging.error('----------------------------------------------------')
            logging.error('Event: {0}'.format(event))
            logging.error('EventType: {0}'.format(event.type()))
            logging.error('Exception error in event loop: {0}'.format(e))
            logging.error('----------------------------------------------------')
            returnValue = False
        finally:
            if isinstance(event, PyQt5.QtGui.QMouseEvent):
                if event.button():
                    logging.debug(f'Mouse: {obj.objectName()}, {event.button()}')
        return returnValue


def except_hook(typeException, valueException, tbackException):
    """
    except_hook implements a wrapper around except hook to log uncatched exceptions to the
    log file. so during user phase I get all the exceptions and logs catched in the file.

    :param typeException:
    :param valueException:
    :param tbackException:
    :return: nothing
    """

    result = traceback.format_exception(typeException, valueException, tbackException)
    logging.error('----------------------------------------------------')
    logging.error('Logging an uncatched Exception')
    logging.error('----------------------------------------------------')
    for i in range(0, len(result)):
        logging.error(result[i].replace('\n', ''))
    logging.error('----------------------------------------------------')
    sys.__excepthook__(typeException, valueException, tbackException)


def setupWorkDirs(mwGlob):
    """
    setupWorkDirs defines the necessary work dirs and checks if they are writable

    :param mwGlob:
    :return: mwGlob
    """

    mwGlob['modeldata'] = mainApp.MountWizzard4.version
    mwGlob['bundleDir'] = ''
    mwGlob['frozen'] = False
    mwGlob['workDir'] = os.getcwd()
    mwGlob['configDir'] = mwGlob['workDir'] + '/config'
    mwGlob['dataDir'] = mwGlob['workDir'] + '/data'
    mwGlob['imageDir'] = mwGlob['workDir'] + '/image'
    mwGlob['tempDir'] = mwGlob['workDir'] + '/temp'
    mwGlob['modelDir'] = mwGlob['workDir'] + '/model'

    for dirPath in ['workDir', 'configDir', 'imageDir', 'dataDir', 'tempDir', 'modelDir']:
        if not os.path.isdir(mwGlob[dirPath]):
            os.makedirs(mwGlob[dirPath])
        if not os.access(mwGlob[dirPath], os.W_OK):
            logging.error('no write access to {0}'.format(dirPath))
    return mwGlob


def checkFrozen():
    """
    checkFrozen extracts data needed to distinguish between real python running setup and
    bundled version of pyinstaller

    :return:
    """

    mwGlob = dict()

    if getattr(sys, 'frozen', False):
        # we are running in a bundle
        # noinspection PyProtectedMember
        mwGlob['bundleDir'] = sys._MEIPASS
        # on mac we have to change path of working directory
        if platform.system() == 'Darwin':
            os.chdir(os.path.dirname(sys.executable))
            os.chdir('..')
            os.chdir('..')
            os.chdir('..')
            mwGlob['frozen'] = True
    else:
        # we are running in a normal Python environment
        mwGlob['bundleDir'] = os.path.dirname(os.path.abspath(__file__))
        mwGlob['frozen'] = False

    return mwGlob


def setupLogging():
    """
    setupLogging defines the logger and formats and disables unnecessary library logging

    :return: true for test purpose
    """
    warnings.filterwarnings("ignore")
    name = 'mw4-{0}.log'.format(datetime.datetime.now().strftime("%Y-%m-%d"))
    logging.basicConfig(level=logging.DEBUG,
                        format='[%(asctime)s.%(msecs)03d]'
                               '[%(levelname)7s]'
                               '[%(filename)20s]'
                               '[%(lineno)5s]'
                               '[%(funcName)20s]'
                               '[%(threadName)10s]'
                               ' > %(message)s',
                        handlers=[logging.FileHandler(name)],
                        datefmt='%Y-%m-%d %H:%M:%S',
                        )

    # setting different log level for imported packages to avoid unnecessary data
    logging.getLogger('mountcontrol').setLevel(logging.DEBUG)
    logging.getLogger('indibase').setLevel(logging.WARNING)
    logging.getLogger('PyQt5').setLevel(logging.ERROR)
    logging.getLogger('requests').setLevel(logging.ERROR)
    # urllib3 is used by requests, so we have to add this as well
    logging.getLogger('urllib3').setLevel(logging.ERROR)
    logging.getLogger('matplotlib').setLevel(logging.ERROR)
    return True


def writeSystemInfo(mwGlob=None):
    """
    writeSystemInfo print overview data to the log file at the beginning of the start

    :return: true for test purpose
    """

    logging.info('## ------------------------------------------------------------------------')
    logging.info('##')
    logging.info('## MountWizzard {0} started !'.format(mwGlob['modeldata']))
    logging.info('##')
    logging.info('## ------------------------------------------------------------------------')
    logging.info('## Platform         : {0}'.format(platform.system()))
    logging.info('## Release          : {0}'.format(platform.release()))
    logging.info('## Machine          : {0}'.format(platform.machine()))
    logging.info('## CPU              : {0}'.format(platform.processor()))
    logging.info('## Python           : {0}'.format(platform.python_version()))
    logging.info('## PyQt5            : {0}'.format(PyQt5.QtCore.PYQT_VERSION_STR))
    logging.info('## Qt               : {0}'.format(PyQt5.QtCore.QT_VERSION_STR))
    logging.info('## Node             : {0}'.format(platform.node()))

    # in some environments I don't get a fully qualified host name
    try:
        hostSummary = socket.gethostbyname_ex(socket.gethostname())
    except socket.herror:
        logging.warning('Could not read properly host configuration')
    except socket.gaierror:
        logging.warning('Could not read properly host configuration')
    else:
        hostsList = hostSummary[2]
        host = [ip for ip in hostsList if not ip.startswith('127.')][: 1]
        for hostname in host:
            logging.info('## IP addr.         : {0}'.format(hostname))
        logging.info('## Hosts            : {0}'.format(hostSummary))

    logging.info('## Environment is   : {0}'.format('frozen' if mwGlob['frozen'] else 'live'))
    logging.info('## Actual workdir   : {0}'.format(mwGlob['workDir']))
    logging.info('## Bundle dir       : {0}'.format(mwGlob['bundleDir']))
    logging.info('## sys.argv[0]      : {0}'.format(sys.argv[0]))
    logging.info('## os.path.basename : {0}'.format(os.path.basename(sys.argv[0])))
    logging.info('## sys.executable   : {0}'.format(sys.executable))

    logging.info('## ------------------------------------------------------------------------')
    logging.info('')
    return True


def extractDataFiles(mwGlob=None, splashW=None):
    """
    we have the necessary files for leap second, ephemerides and satellite already stored
    in the files system of the app. we bring them to the file system of the user, if they
    do not exist.

    :return: True fpr test purpose
    """

    if mwGlob is None:
        return False

    files = [
        'Leap_Second.dat',
        'deltat.data',
        'deltat.preds',
        'de421_23.bsp',
        'active.txt',
    ]
    for file in files:
        if splashW is not None:
            splashW.showMessage('Loading {0}'.format(file))
        filePath = mwGlob['dataDir'] + '/' + file
        if os.path.isfile(filePath):
            continue
        # as we cannot access data from Qt resource system, we have to convert it to
        # ByteIO first
        stream = PyQt5.QtCore.QFile(f':/{file}')
        stream.open(PyQt5.QtCore.QFile.ReadOnly)
        with open(filePath, 'wb') as outFile:
            outFile.write(stream.readAll())
        stream.close()
    return True


def main():
    """
    main prepares the loading of mountwizzard application. it prepares a splash screen
    and handler the setup of the logger, bundle handling etc. in addition some information
    about the system are written into the logfile to be able to debug in different conditions
    the system environment.

    :return: nothing
    """

    # initiating the main app
    app = MyApp(sys.argv)
    # app = PyQt5.QtWidgets.QApplication(sys.argv)

    # generating splash screen
    splashW = splash.SplashScreen(application=app)

    # and start with a first splash screen
    splashW.showMessage('Start initialising')
    splashW.setValue(0)

    # checking workdir and if the system is started from frozen app
    mwGlob = checkFrozen()
    mwGlob = setupWorkDirs(mwGlob)

    # now setup the logging environment
    splashW.showMessage('Setup logging')
    splashW.setValue(20)
    setupLogging()

    # start logging with basic system data for information
    splashW.showMessage('Write system info to log')
    splashW.setValue(40)
    writeSystemInfo(mwGlob=mwGlob)

    # loading leap seconds, spice kernel and hipparcos catalogue
    splashW.showMessage('Loading star and time data')
    splashW.setValue(60)
    extractDataFiles(mwGlob=mwGlob, splashW=splashW)

    # and finally starting the application
    splashW.showMessage('Loading Data')
    splashW.setValue(80)
    sys.excepthook = except_hook
    app.setWindowIcon(PyQt5.QtGui.QIcon(':/mw4.ico'))
    mountApp = mainApp.MountWizzard4(mwGlob)
    mountApp.mainW.show()

    # end of splash screen
    splashW.showMessage('Finishing loading')
    splashW.setValue(100)
    splashW.close()

    # quit app
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
